import threading
import typing
import base64
from datetime import datetime
from sqlalchemy import create_engine, select, inspect as inspect_database
from sqlalchemy import String, JSON, LargeBinary, Integer
from sqlalchemy.ext import asyncio as asyncio_ext
from sqlalchemy.orm import sessionmaker, Mapped, mapped_column, DeclarativeBase, MappedAsDataclass
from sqlite3 import DatabaseError
from pymongo import MongoClient, errors as mongo_errors
from dataclasses import dataclass, asdict

from ..param import Parameterized
from ..core.property import Property
from ..constants import JSONSerializable
from ..serializers.serializers import PythonBuiltinJSONSerializer as JSONSerializer, BaseSerializer, Serializers
from .config_models import MongoDBConfig, SQLDBConfig, SQLiteConfig


class ThingTableBase(DeclarativeBase):
    """SQLAlchemy base table for all thing related tables"""

    pass


class SerializedProperty(MappedAsDataclass, ThingTableBase):
    """
    Property value is serialized before storing in database, therefore providing unified version for
    SQLite and other relational tables
    """

    __tablename__ = "properties"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    serialized_value: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)
    thing_id: Mapped[str] = mapped_column(String)
    thing_class: Mapped[str] = mapped_column(String)
    created_at: Mapped[str] = mapped_column(String)
    updated_at: Mapped[str] = mapped_column(String)
    content_type: Mapped[str] = mapped_column(String, nullable=False, default="application/json")


class ThingInformation(MappedAsDataclass, ThingTableBase):
    """Stores information about the Thing instance itself, useful metadata which may be later populated in a GUI"""

    __tablename__ = "things"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    thing_id: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    thing_class: Mapped[str] = mapped_column(String, nullable=False)
    script: Mapped[str] = mapped_column(String)
    init_kwargs: Mapped[JSONSerializable] = mapped_column(JSON)
    server_id: Mapped[str] = mapped_column(String)

    def json(self):
        return asdict(self)


@dataclass
class DeserializedProperty:  # not part of database
    """Property with deserialized value after fetching from database"""

    thing_id: str
    name: str
    value: typing.Any
    created_at: str
    updated_at: str


class BaseDB:
    """Implements configuration file reader for all irrespective sync or async DB operation"""

    def __init__(self, instance: Parameterized, config_file: typing.Union[str, None] = None) -> None:
        self.thing_instance = instance
        self.conf = BaseDB.load_conf(config_file, default_file_path=f"{instance.id}.db")
        self.URL = self.conf.URL
        self._batch_call_context = {}

    @classmethod
    def load_conf(
        cls,
        config_file: str,
        default_file_path: str = "",
    ) -> typing.Union[SQLDBConfig, SQLiteConfig, MongoDBConfig]:
        """
        load configuration file using JSON serializer
        """
        if not config_file:
            return SQLiteConfig(file=default_file_path)
        elif config_file.endswith(".json"):
            file = open(config_file, "r")
            conf = JSONSerializer.load(file)
            if conf.get("provider", None) in ["postgresql", "mysql"]:
                return SQLDBConfig.model_validate(conf, strict=True, from_attributes=True)
            elif conf.get("provider", None) == "sqlite":
                return SQLiteConfig.model_validate(conf, strict=True, from_attributes=True)
            elif conf.get("provider", None) == "mongo":
                return MongoDBConfig.model_validate(conf, strict=True, from_attributes=True)
            raise NotImplementedError("only postgresql, mysql, sqlite and mongo are supported")
        raise ValueError("config files of extension {} expected, given file name {}".format(["json"], config_file))

    @property
    def in_batch_call_context(self):
        return threading.get_ident() in self._batch_call_context


class BaseAsyncDB(BaseDB):
    """
    Base class for an async database engine, implements configuration file reader,
    sqlalchemy engine & session creation. Set `async_db_engine` boolean flag to True in `Thing` class
    to use this engine. Database operations are then scheduled in the event loop instead of blocking the current thread.
    Scheduling happens after properties are set/written.

    Parameters
    ----------
    database: str
        The database to open in the database server specified in config_file (see below)
    serializer: BaseSerializer
        The serializer to use for serializing and deserializing data (for example
        property serializing before writing to database). Will be the same as zmq_serializer supplied to `Thing`.
    config_file: str
        absolute path to database server configuration file
    """

    def __init__(
        self,
        instance: Parameterized,
        serializer: typing.Optional[BaseSerializer] = None,
        config_file: typing.Union[str, None] = None,
    ) -> None:
        super().__init__(instance=instance, serializer=serializer, config_file=config_file)
        self.engine = asyncio_ext.create_async_engine(self.URL)
        self.async_session = sessionmaker(self.engine, expire_on_commit=True, class_=asyncio_ext.AsyncSession)
        if self.conf.provider == "sqlite":
            ThingTableBase.metadata.create_all(self.engine)


class BaseSyncDB(BaseDB):
    """
    Base class for an synchronous (blocking) database engine, implements configuration file reader, sqlalchemy engine
    & session creation. Default DB engine for `Thing` & called immediately after properties are set/written.

    Parameters
    ----------
    database: str
        The database to open in the database server specified in config_file (see below)
    serializer: BaseSerializer
        The serializer to use for serializing and deserializing data (for example
        property serializing into database for storage). Will be the same as
        zmq_serializer supplied to `Thing`.
    config_file: str
        absolute path to database server configuration file
    """

    def __init__(self, instance: Parameterized, config_file: typing.Union[str, None] = None) -> None:
        super().__init__(instance=instance, config_file=config_file)
        self.engine = create_engine(self.URL)
        self.sync_session = sessionmaker(self.engine, expire_on_commit=True)
        if self.conf.provider == "sqlite":
            ThingTableBase.metadata.create_all(self.engine)


class ThingDB(BaseSyncDB):
    """
    Database engine composed within `Thing`, carries out database operations like storing object information, properties
    etc.

    Parameters
    ----------
    id: str
        `id` of the `Thing`
    config_file: str
        configuration file of the database server
    """

    def fetch_own_info(self):  # -> ThingInformation:
        """
        fetch `Thing` instance's own information (some useful metadata which helps the `Thing` run).

        Returns
        -------
        `ThingInformation`
        """
        if not inspect_database(self.engine).has_table("things"):
            return
        with self.sync_session() as session:
            stmt = select(ThingInformation).filter_by(
                thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__
            )
            data = session.execute(stmt)
            data = data.scalars().all()
            if len(data) == 0:
                return None
            elif len(data) == 1:
                return data[0]
            else:
                raise DatabaseError(
                    "Multiple things with same instance name found, either cleanup database/detach/make new"
                )

    def get_property(self, property: typing.Union[str, Property], deserialized: bool = True) -> typing.Any:
        """
        fetch a single property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        deserialized: bool, default True
            deserialize the property if True

        Returns
        -------
        value: Any
            property value
        """
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id,
                thing_class=self.thing_instance.__class__.__name__,
                name=name,
            )
            data = session.execute(stmt)
            prop = data.scalars().all()  # type: typing.Sequence[SerializedProperty]
            if len(prop) == 0:
                raise DatabaseError(f"property {name} not found in database")
            elif len(prop) > 1:
                raise DatabaseError("multiple properties with same name found")  # Impossible actually
            if not deserialized:
                return prop[0]
            serializer = Serializers.content_types.get(prop[0].content_type, None) or Serializers.for_object(
                self.thing_instance.id, self.thing_instance.__class__.__name__, name
            )
            return serializer.loads(prop[0].serialized_value)

    def set_property(self, property: typing.Union[str, Property], value: typing.Any) -> None:
        """
        change the value of an already existing property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        value: Any
            value of the property
        """
        if self.in_batch_call_context:
            self._batch_call_context[threading.get_ident()][property.name] = value
            return
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id,
                thing_class=self.thing_instance.__class__.__name__,
                name=name,
            )
            data = session.execute(stmt)
            prop = data.scalars().all()
            if len(prop) > 1:
                raise DatabaseError("multiple properties with same name found")  # Impossible actually
            if len(prop) == 1:
                prop = prop[0]
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, name
                )
                prop.serialized_value = serializer.dumps(value)
                prop.updated_at = datetime.now().isoformat()
                prop.content_type = serializer.content_type
            else:
                serializer = Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, name
                )
                now = datetime.now().isoformat()
                prop = SerializedProperty(
                    id=None,
                    name=name,
                    serialized_value=serializer.dumps(value),
                    thing_id=self.thing_instance.id,
                    thing_class=self.thing_instance.__class__.__name__,
                    created_at=now,
                    updated_at=now,
                    content_type=serializer.content_type,
                )
                session.add(prop)
            session.commit()

    def get_properties(
        self, properties: typing.Dict[typing.Union[str, Property], typing.Any], deserialized: bool = True
    ) -> typing.Dict[str, typing.Any]:
        """
        get multiple properties at once.

        Parameters
        ----------
        properties: List[str | Property]
            string names or the descriptor of the properties as a list

        Returns
        -------
        value: Dict[str, Any]
            property names and values as items
        """
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = (
                select(SerializedProperty)
                .filter_by(thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__)
                .filter(SerializedProperty.name.in_(names))
            )
            data = session.execute(stmt)
            unserialized_props = data.scalars().all()
            props = dict()
            for prop in unserialized_props:
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, prop.name
                )
                props[prop.name] = (
                    prop.serialized_value if not deserialized else serializer.loads(prop.serialized_value)
                )
            return props

    def set_properties(self, properties: typing.Dict[typing.Union[str, Property], typing.Any]) -> None:
        """
        change the values of already existing few properties at once

        Parameters
        ----------
        properties: Dict[str | Property, Any]
            string names or the descriptor of the property and any value as dictionary pairs
        """
        if self.in_batch_call_context:
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                self._batch_call_context[threading.get_ident()][name] = value
            return
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = (
                select(SerializedProperty)
                .filter_by(thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__)
                .filter(SerializedProperty.name.in_(names))
            )
            data = session.execute(stmt)
            db_props = data.scalars().all()
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                db_prop = list(filter(lambda db_prop: db_prop.name == name, db_props))  # type: typing.List[SerializedProperty]
                if len(db_prop) > 1:
                    raise DatabaseError("multiple properties with same name found")  # Impossible actually
                if len(db_prop) == 1:
                    db_prop = db_prop[0]  # type: SerializedProperty
                    serializer = Serializers.content_types.get(db_prop.content_type, None) or Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, name
                    )
                    db_prop.serialized_value = serializer.dumps(value)
                    db_prop.updated_at = datetime.now().isoformat()
                    db_prop.content_type = serializer.content_type
                else:
                    serializer = Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, name
                    )
                    now = datetime.now().isoformat()
                    prop = SerializedProperty(
                        id=None,
                        name=name,
                        serialized_value=serializer.dumps(value),
                        thing_id=self.thing_instance.id,
                        thing_class=self.thing_instance.__class__.__name__,
                        created_at=now,
                        updated_at=now,
                        content_type=serializer.content_type,
                    )
                    session.add(prop)
            session.commit()

    def get_all_properties(self, deserialized: bool = True) -> typing.Dict[str, typing.Any]:
        """
        read all properties of the `Thing` instance.

        Parameters
        ----------
        deserialized: bool, default True
            deserilize the properties if True
        """
        with self.sync_session() as session:
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__
            )
            data = session.execute(stmt)
            existing_props = data.scalars().all()  # type: typing.Sequence[SerializedProperty]
            if not deserialized:
                return existing_props
            props = dict()
            for prop in existing_props:
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, prop.name
                )
                props[prop.name] = serializer.loads(prop.serialized_value)
            return props

    def create_missing_properties(
        self, properties: typing.Dict[str, Property], get_missing_property_names: bool = False
    ) -> None:
        """
        create any and all missing properties of `Thing` instance in database.

        Parameters
        ----------
        properties: Dict[str, Property]
            descriptors of the properties

        Returns
        -------
        List[str]
            list of missing properties if get_missing_property_names is True
        """
        missing_props = []
        with self.sync_session() as session:
            existing_props = self.get_all_properties()
            for prop in properties.values():
                if prop.name not in existing_props:
                    serializer = Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, prop.name
                    )
                    now = datetime.now().isoformat()
                    prop = SerializedProperty(
                        id=None,
                        name=prop.name,
                        serialized_value=serializer.dumps(getattr(self.thing_instance, prop.name)),
                        thing_id=self.thing_instance.id,
                        thing_class=self.thing_instance.__class__.__name__,
                        created_at=now,
                        updated_at=now,
                        content_type=serializer.content_type,
                    )
                    session.add(prop)
                    missing_props.append(prop.name)
            session.commit()
        if get_missing_property_names:
            return missing_props

    def create_db_init_properties(
        self, thing_id: str = None, thing_class: str = None, **properties: typing.Any
    ) -> None:
        """
        Create properties that are initialized from database for a thing instance.
        Invoke this method once before running the thing instance to store its initial value in database.

        Parameters
        ----------
        thing_id: str
            ID of the thing instance to which these properties belong
        thing_class: str
            Class name of the thing instance to which these properties belong
        properties: Dict[str, Any]
            property names and their initial values as dictionary pairs
        """
        with self.sync_session() as session:
            for name, value in properties.items():
                serializer = Serializers.for_object(thing_id, thing_class, name)
                now = datetime.now().isoformat()
                prop = SerializedProperty(
                    id=None,
                    name=name,
                    serialized_value=serializer.dumps(value),
                    thing_id=thing_id,
                    thing_class=thing_class,
                    created_at=now,
                    updated_at=now,
                    content_type=serializer.content_type,
                )
                session.add(prop)
            session.commit()


class batch_db_commit:
    """
    Context manager to write multiple properties to database at once. Useful for sequential sets/writes of multiple properties
    which has db_commit or db_persist set to True, but only write their values to database at once.
    """

    def __init__(self, db_engine: ThingDB) -> None:
        self.db_engine = db_engine

    def __enter__(self) -> None:
        self.db_engine._context[threading.get_ident()] = dict()

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        data = self.db_engine._context.pop(threading.get_ident(), dict())  # typing.Dict[str, typing.Any]
        if exc_type is None:
            self.db_engine.set_properties(data)
            return
        for name, value in data.items():
            try:
                self.db_engine.set_property(name, value)
            except Exception as ex:
                self.db_engine.thing_instance.logger.error(
                    f"failed to set property {name} to value {value} during batch commit due to exception {ex}"
                )


class MongoThingDB:
    """
    MongoDB-backed database engine for Thing properties and info.

    This class provides persistence for Thing properties using MongoDB.
    Properties are stored in the 'properties' collection, with fields:
    - id: Thing instance identifier
    - name: property name
    - serialized_value: serialized property value

    Methods mirror the interface of ThingDB for compatibility.
    """

    def __init__(self, instance: Parameterized, config_file: typing.Union[str, None] = None) -> None:
        """
        Initialize MongoThingDB for a Thing instance.
        Connects to MongoDB and sets up collections.
        """
        self.thing_instance = instance
        self.id = instance.id
        self.config = self.load_conf(config_file)
        self.client = MongoClient(self.config.get("mongo_uri", "mongodb://localhost:27017"))
        self.db = self.client[self.config.get("database", "hololinked")]
        self.properties = self.db["properties"]
        self.things = self.db["things"]

    @classmethod
    def load_conf(cls, config_file: str) -> typing.Dict[str, typing.Any]:
        """
        Load configuration from JSON file if provided.
        """
        if not config_file:
            return {}
        elif config_file.endswith(".json"):
            with open(config_file, "r") as file:
                return JSONSerializer.load(file)
        else:
            raise ValueError(f"config files of extension - ['json'] expected, given file name {config_file}")

    def fetch_own_info(self):
        """
        Fetch Thing instance metadata from the 'things' collection.
        """
        doc = self.things.find_one({"id": self.id})
        return doc

    def get_property(self, property: typing.Union[str, Property], deserialized: bool = True) -> typing.Any:
        """
        Get a property value from MongoDB for this Thing.
        If deserialized=True, returns the Python value.
        """
        name = property if isinstance(property, str) else property.name
        doc = self.properties.find_one({"id": self.id, "name": name})
        if not doc:
            raise mongo_errors.PyMongoError(f"property {name} not found in database")
        if not deserialized:
            return doc
        serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
        return serializer.loads(base64.b64decode(doc["serialized_value"]))

    def set_property(self, property: typing.Union[str, Property], value: typing.Any) -> None:
        """
        Set a property value in MongoDB for this Thing.
        Value is serialized before storage.
        """
        name = property if isinstance(property, str) else property.name
        serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
        serialized_value = base64.b64encode(serializer.dumps(value)).decode("utf-8")
        self.properties.update_one(
            {"id": self.id, "name": name}, {"$set": {"serialized_value": serialized_value}}, upsert=True
        )

    def get_properties(
        self, properties: typing.Dict[typing.Union[str, Property], typing.Any], deserialized: bool = True
    ) -> typing.Dict[str, typing.Any]:
        """
        Get multiple property values from MongoDB for this Thing.
        Returns a dict of property names to values.
        """
        names = [obj if isinstance(obj, str) else obj.name for obj in properties.keys()]
        cursor = self.properties.find({"id": self.id, "name": {"$in": names}})
        result = {}
        for doc in cursor:
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, doc["name"])
            result[doc["name"]] = (
                doc["serialized_value"]
                if not deserialized
                else serializer.loads(base64.b64decode(doc["serialized_value"]))
            )
        return result

    def set_properties(self, properties: typing.Dict[typing.Union[str, Property], typing.Any]) -> None:
        """
        Set multiple property values in MongoDB for this Thing.
        """
        for obj, value in properties.items():
            name = obj if isinstance(obj, str) else obj.name
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
            serialized_value = base64.b64encode(serializer.dumps(value)).decode("utf-8")
            self.properties.update_one(
                {"id": self.id, "name": name}, {"$set": {"serialized_value": serialized_value}}, upsert=True
            )

    def get_all_properties(self, deserialized: bool = True) -> typing.Dict[str, typing.Any]:
        cursor = self.properties.find({"id": self.id})
        result = {}
        for doc in cursor:
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, doc["name"])
            result[doc["name"]] = (
                doc["serialized_value"]
                if not deserialized
                else serializer.loads(base64.b64decode(doc["serialized_value"]))
            )
        return result

    def create_missing_properties(
        self, properties: typing.Dict[str, Property], get_missing_property_names: bool = False
    ) -> typing.Any:
        missing_props = []
        existing_props = self.get_all_properties()
        for name, new_prop in properties.items():
            if name not in existing_props:
                serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, new_prop.name)
                serialized_value = base64.b64encode(
                    serializer.dumps(getattr(self.thing_instance, new_prop.name))
                ).decode("utf-8")
                self.properties.insert_one({"id": self.id, "name": new_prop.name, "serialized_value": serialized_value})
                missing_props.append(name)
        if get_missing_property_names:
            return missing_props


__all__ = [BaseAsyncDB.__name__, BaseSyncDB.__name__, ThingDB.__name__, batch_db_commit.__name__]



import typing
from dataclasses import dataclass

from ..constants import byte_types
from .serializers import Serializers, BaseSerializer


@dataclass
class SerializableData:
    """
    A container for data that can be serialized.
    The content type decides the serializer to be used.
    """

    value: typing.Any
    serializer: BaseSerializer | None = None
    content_type: str = "application/json"
    _serialized: bytes | None = None

    def serialize(self):
        """serialize the value"""
        if self._serialized is not None:
            return self._serialized
        if isinstance(self.value, byte_types):
            return self.value
        if self.serializer is not None:
            return self.serializer.dumps(self.value)
        serializer = Serializers.content_types.get(self.content_type, None)
        if serializer is not None:
            return serializer.dumps(self.value)
        raise ValueError(f"content type {self.content_type} not supported for serialization")

    def deserialize(self):
        """deserialize the value"""
        if not isinstance(self.value, byte_types):
            return self.value
        if self.serializer is not None:
            return self.serializer.loads(self.value)
        serializer = Serializers.content_types.get(self.content_type, None)
        if serializer is not None:
            return serializer.loads(self.value)
        raise ValueError(f"content type {self.content_type} not supported for deserialization")

    def require_serialized(self) -> None:
        """ensure the value is serialized"""
        self._serialized = self.serialize()


@dataclass
class PreserializedData:
    """
    A container for data that is already serialized.
    The content type may indicate the serializer used.
    """

    value: bytes
    content_type: str = "application/octet-stream"



import asyncio
import base64
import random
import uuid
import unittest
import time
import logging
import requests
from typing import Any
from dataclasses import dataclass
from types import SimpleNamespace

from hololinked.client.abstractions import SSE
from hololinked.utils import pep8_to_dashed_name
from hololinked.config import global_config
from hololinked.constants import ZMQ_TRANSPORTS
from hololinked.core.zmq.message import (
    ServerExecutionContext,
    ThingExecutionContext,
    default_server_execution_context,
)
from hololinked.serializers import JSONSerializer
from hololinked.serializers.payloads import PreserializedData, SerializableData
from hololinked.serializers.serializers import (
    MsgpackSerializer,
    PickleSerializer,
    BaseSerializer,
)
from hololinked.core.meta import ThingMeta
from hololinked.core.zmq.rpc_server import (
    RPCServer,
)  # sets loop policy, TODO: move somewhere else
from hololinked.client.proxy import ObjectProxy
from hololinked.client.factory import ClientFactory
from hololinked.server.http import HTTPServer
from hololinked.server.http.handlers import (
    PropertyHandler,
    RPCHandler,
    ThingDescriptionHandler,
)
from hololinked.server.security import Argon2BasicSecurity, BcryptBasicSecurity
from hololinked.td.security_definitions import SecurityScheme

try:
    from .things import OceanOpticsSpectrometer, TestThing
    from .utils import TestCase, TestRunner, fake
    from .test_11_rpc_e2e import TestRPCEndToEnd
except ImportError:
    from things import OceanOpticsSpectrometer, TestThing
    from utils import TestCase, TestRunner, fake
    from test_11_rpc_e2e import TestRPCEndToEnd


class TestHTTPServer(TestCase):
    def test_01_init_run_and_stop(self):
        """Test basic init, run and stop of the HTTP server."""
        # init, run and stop synchronously
        server = HTTPServer(log_level=logging.ERROR + 10, port=60001)
        self.assertTrue(server.all_ok)
        server.listen(forked=True)
        time.sleep(5)
        server.stop()
        time.sleep(2)

        # stop remotely
        server.listen(forked=True)
        time.sleep(5)
        response = requests.post(f"http://127.0.0.1:{server.port}/stop")
        self.assertIn(response.status_code, [200, 201, 202, 204])
        time.sleep(2)

    def test_02_add_interaction_affordance(self):
        """Test adding an interaction affordance to the HTTP server."""
        server = HTTPServer(log_level=logging.ERROR + 10)
        self.assertTrue(server.all_ok)

        # add an interaction affordance
        server.add_property("/max-intensity", OceanOpticsSpectrometer.max_intensity)
        server.add_action("/connect", OceanOpticsSpectrometer.connect)
        server.add_event("/intensity/event", OceanOpticsSpectrometer.intensity_measurement_event)

        self.assertIn("/max-intensity", server.router)
        self.assertIn("/connect", server.router)
        self.assertIn("/intensity/event", server.router)

        # replacing interation affordances on an existing URL path causes a warning
        self.assertWarns(
            UserWarning,
            server.add_property,
            "/max-intensity",
            OceanOpticsSpectrometer.last_intensity,
        )
        self.assertWarns(
            UserWarning,
            server.add_action,
            "/connect",
            OceanOpticsSpectrometer.disconnect,
        )
        self.assertWarns(
            UserWarning,
            server.add_event,
            "/intensity/event",
            OceanOpticsSpectrometer.intensity_measurement_event,
        )

    def test_03_add_thing(self):
        """Test adding a Thing object to the HTTP server."""

        # add a thing, both class and instance
        server = HTTPServer(log_level=logging.ERROR + 10)
        for thing in [
            OceanOpticsSpectrometer(id="test", log_level=logging.ERROR + 10),
            TestThing(id="test-thing", log_level=logging.ERROR + 10),
        ]:
            old_number_of_rules = len(server.app.wildcard_router.rules) + len(server.router._pending_rules)
            server.add_things(thing)
            # TODO - reinstate rule numbers as they ensure that all routes were added
            # self.assertTrue(
            #     len(server.app.wildcard_router.rules) + len(server.router._pending_rules) - old_number_of_rules >=
            #     len(thing.properties.remote_objects) + len(thing.actions) + len(thing.events)
            # )
            # server.router.print_rules()

        old_number_of_rules = len(server.app.wildcard_router.rules) + len(server.router._pending_rules)
        for thing_meta in [OceanOpticsSpectrometer, TestThing]:
            self.assertRaises(ValueError, server.add_things, thing_meta)
        self.assertTrue(
            len(server.app.wildcard_router.rules) + len(server.router._pending_rules) == old_number_of_rules
        )

        # create new server to compute number of rules
        server = HTTPServer(log_level=logging.ERROR + 10)
        thing = OceanOpticsSpectrometer(id="test", log_level=logging.ERROR + 10)
        old_number_of_rules = len(server.app.wildcard_router.rules) + len(server.router._pending_rules)
        # append route with /custom to denote its a custom route
        server.add_property("/max-intensity/custom", OceanOpticsSpectrometer.max_intensity)
        server.add_action("/connect/custom", OceanOpticsSpectrometer.connect)
        server.add_event("/intensity/event/custom", OceanOpticsSpectrometer.intensity_measurement_event)
        server.add_things(thing)
        self.assertIn(f"/{thing.id}/max-intensity/custom", server.router)
        self.assertIn(f"/{thing.id}/connect/custom", server.router)
        self.assertIn(f"/{thing.id}/intensity/event/custom", server.router)
        # check if the affordance was not added twice using the default paths while add_thing was called
        self.assertNotIn(
            f"/{pep8_to_dashed_name(OceanOpticsSpectrometer.max_intensity.name)}",
            server.router,
        )
        self.assertNotIn(
            f"/{pep8_to_dashed_name(OceanOpticsSpectrometer.connect.name)}",
            server.router,
        )
        self.assertNotIn(
            f"/{pep8_to_dashed_name(OceanOpticsSpectrometer.intensity_measurement_event.name)}",
            server.router,
        )
        # TODO - reinstate rule numbers as they ensure that all routes were added
        # self.assertTrue(
        #         len(server.app.wildcard_router.rules) + len(server.router._pending_rules) - old_number_of_rules >=
        #         len(thing.properties.remote_objects) + len(thing.actions) + len(thing.events)
        #     )
        # also check that it does not create duplicate rules

    def test_04_add_thing_over_zmq_server(self):
        """extension of previous two tests to complete adding a thing running over a zmq server"""
        server = HTTPServer(log_level=logging.ERROR + 10)
        old_number_of_rules = len(server.app.wildcard_router.rules) + len(server.router._pending_rules)

        thing_id = f"test-add-zmq-{uuid.uuid4().hex[0:8]}"
        thing = OceanOpticsSpectrometer(id=thing_id, log_level=logging.ERROR + 10)
        thing.run_with_zmq_server(ZMQ_TRANSPORTS.INPROC, forked=True)

        server.add_property("/max-intensity/custom", OceanOpticsSpectrometer.max_intensity)
        server.add_action("/connect/custom", OceanOpticsSpectrometer.connect)
        server.add_event(
            "/intensity/event/custom",
            OceanOpticsSpectrometer.intensity_measurement_event,
        )
        server.add_things(thing)

        # server.router.print_rules()
        # print(thing.properties.remote_objects.keys(), thing.actions.descriptors.keys(), thing.events.descriptors.keys())
        # self.assertTrue(
        #     len(server.app.wildcard_router.rules) + len(server.router._pending_rules) - old_number_of_rules >=
        #     len(thing.properties.remote_objects) + len(thing.actions) + len(thing.events)
        # )

        fake_request = SimpleNamespace(path=f"/{thing_id}/max-intensity/custom")
        self.assertTrue(
            any([rule.matcher.match(fake_request) is not None for rule in server.app.wildcard_router.rules])
        )
        fake_request = SimpleNamespace(path="/non-existing-path-that-i-know-will-not-match")
        self.assertFalse(
            any([rule.matcher.match(fake_request) is not None for rule in server.app.wildcard_router.rules])
        )
        fake_request = SimpleNamespace(path=f"/{thing_id}/connect/custom")
        self.assertTrue(
            any([rule.matcher.match(fake_request) is not None for rule in server.app.wildcard_router.rules])
        )
        fake_request = SimpleNamespace(path=f"/{thing_id}/intensity/event/custom")
        self.assertTrue(
            any([rule.matcher.match(fake_request) is not None for rule in server.app.wildcard_router.rules])
        )

        while not thing.rpc_server:
            time.sleep(0.1)  # wait for rpc server to be ready
        thing.rpc_server.stop()

    def test_05_handlers(self):
        """Test request info and payload decoding in RPC handlers along with content type handling"""
        latest_request_info = None  # type: "LatestRequestInfo"

        @dataclass
        class LatestRequestInfo:
            server_execution_context: ServerExecutionContext | dict[str, Any]
            thing_execution_context: ThingExecutionContext | dict[str, Any]
            payload: SerializableData
            preserialized_payload: PreserializedData

        class TestableRPCHandler(RPCHandler):
            def update_latest_request_info(self) -> None:
                nonlocal latest_request_info
                server_execution_context, thing_execution_context, _, _ = self.get_execution_parameters()
                payload, preserialized_payload = self.get_request_payload()
                latest_request_info = LatestRequestInfo(
                    server_execution_context=server_execution_context,
                    thing_execution_context=thing_execution_context,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                )

            async def get(self):
                self.update_latest_request_info()
                self.set_status(200)
                self.finish()

            async def put(self):
                self.update_latest_request_info()
                self.set_status(200)
                self.finish()

            async def post(self):
                # for exit to go through
                await self.handle_through_thing("invokeaction")

        global_config.ALLOW_PICKLE = True  # allow pickle serializer for testing
        thing_id = f"test-request-info-{uuid.uuid4().hex[0:8]}"
        port = 60002

        thing = OceanOpticsSpectrometer(id=thing_id, log_level=logging.ERROR + 10)
        thing.run_with_http_server(
            port=port,
            forked=True,
            property_handler=TestableRPCHandler,
            action_handler=TestableRPCHandler,
        )
        self.wait_until_server_ready(port=port)
        session = requests.session()
        for serializer in [JSONSerializer(), MsgpackSerializer(), PickleSerializer()]:
            serializer: BaseSerializer
            for method, path, body in [
                # server and thing execution context tests
                ("get", f"/{thing_id}/integration-time", None),
                ("get", f"/{thing_id}/integration-time?fetchExecutionLogs=true", None),
                (
                    "get",
                    f"/{thing_id}/integration-time?fetchExecutionLogs=true&oneway=true",
                    None,
                ),
                (
                    "get",
                    f"/{thing_id}/integration-time?oneway=true&invokationTimeout=100",
                    None,
                ),
                (
                    "get",
                    f"/{thing_id}/integration-time?invokationTimeout=100&executionTimeout=120&fetchExecutionLogs=true",
                    None,
                ),
                # test payloads for JSON content type
                ("put", f"/{thing_id}/integration-time", 1200),
                (
                    "put",
                    f"/{thing_id}/integration-time?fetchExecutionLogs=true",
                    {"a": 1, "b": 2},
                ),
                (
                    "put",
                    f"/{thing_id}/integration-time?fetchExecutionLogs=true&oneway=true",
                    [1, 2, 3],
                ),
                (
                    "put",
                    f"/{thing_id}/integration-time?oneway=true&invokationTimeout=100",
                    "abcd",
                ),
                (
                    "put",
                    f"/{thing_id}/integration-time?invokationTimeout=100&executionTimeout=120&fetchExecutionLogs=true",
                    True,
                ),
                # test payloads for other content types
            ]:
                response = session.request(
                    method=method,
                    url=f"http://127.0.0.1:{port}{path}",
                    data=serializer.dumps(body) if body is not None else None,
                    headers={"Content-Type": serializer.content_type},
                )
                self.assertTrue(response.status_code in [200, 201, 202, 204])
                assert isinstance(latest_request_info, LatestRequestInfo)
                # test ThingExecutionContext
                self.assertTrue(
                    isinstance(
                        latest_request_info.thing_execution_context,
                        ThingExecutionContext,
                    )
                )
                self.assertTrue(
                    ("fetchExecutionLogs" in path and latest_request_info.thing_execution_context.fetchExecutionLogs)
                    or not latest_request_info.thing_execution_context.fetchExecutionLogs
                )
                # test ServerExecutionContext
                self.assertTrue(
                    isinstance(
                        latest_request_info.server_execution_context,
                        ServerExecutionContext,
                    )
                )
                self.assertTrue(
                    ("oneway" in path and latest_request_info.server_execution_context.oneway)
                    or not latest_request_info.server_execution_context.oneway
                )
                self.assertTrue(
                    (
                        "invokationTimeout" in path
                        and latest_request_info.server_execution_context.invokationTimeout == 100
                    )
                    or
                    # assume that in all tests where invokation timeout is specified, it will be 100
                    latest_request_info.server_execution_context.invokationTimeout
                    == default_server_execution_context.invokationTimeout
                )
                self.assertTrue(
                    (
                        "executionTimeout" in path
                        and latest_request_info.server_execution_context.executionTimeout == 120
                    )
                    or
                    # assume that in all tests where execution timeout is specified, it will be 120
                    latest_request_info.server_execution_context.executionTimeout
                    == default_server_execution_context.executionTimeout
                )
                # test body
                self.assertTrue(latest_request_info.payload.deserialize() == body)

        self.stop_server(port=port, thing_ids=[thing_id])

    def _test_handlers_end_to_end(self, port: int, thing_id: str, **request_kwargs):
        """
        basic end-to-end test with the HTTP server using handlers.
        Auth & other features not included, only invokation of interaction affordances.
        """
        session = requests.Session()
        logging.getLogger("requests").setLevel(logging.CRITICAL)
        logging.getLogger("urllib3").setLevel(logging.CRITICAL)
        # test end to end
        for method, path, body in self.generate_endpoints_for_thing(OceanOpticsSpectrometer, thing_id):
            # request will go through the Thing object
            response = session.request(
                method=method,
                url=f"http://127.0.0.1:{port}{path}",
                data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
                **request_kwargs,
            )
            self.assertTrue(response.status_code in [200, 201, 202, 204])
            # check if the response body is as expected
            if body and method != "put":
                self.assertTrue(response.json() == body)
            # check headers
            self.assertIn("Access-Control-Allow-Origin", response.headers)
            self.assertIn("Access-Control-Allow-Credentials", response.headers)
            self.assertIn("Content-Type", response.headers)

        # test unsupported HTTP methods
        for method, path, body in self.generate_endpoints_for_thing(OceanOpticsSpectrometer, thing_id):
            response = session.request(
                method="post"
                if method in ["get", "put"]
                else random.choice(["put", "delete"])
                if method == "post"
                else method,
                # get and put become post and post becomes put
                # i.e swap the default HTTP method with an unsupported one to generate 405
                url=f"http://127.0.0.1:{port}{path}",
                data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
                **request_kwargs,
            )
            self.assertTrue(response.status_code == 405)

        # check options for supported HTTP methods
        for method, path, body in self.generate_endpoints_for_thing(OceanOpticsSpectrometer, thing_id):
            response = session.options(f"http://127.0.0.1:{port}{path}", **request_kwargs)
            self.assertTrue(response.status_code in [200, 201, 202, 204])
            self.assertIn("Access-Control-Allow-Origin", response.headers)
            self.assertIn("Access-Control-Allow-Credentials", response.headers)
            self.assertIn("Access-Control-Allow-Headers", response.headers)
            self.assertIn("Access-Control-Allow-Methods", response.headers)
            allow_methods = response.headers.get("Access-Control-Allow-Methods", [])
            self.assertTrue(
                method.upper() in allow_methods,
                f"Method {method} not allowed in {allow_methods}",
            )

    def _test_invalid_auth_end_to_end(self, port: int, thing_id: str, wrong_auth_headers: list[str] = None):
        # check wrong credentials
        session = requests.Session()
        for wrong_auth in wrong_auth_headers:
            for method, path, body in self.generate_endpoints_for_thing(OceanOpticsSpectrometer, thing_id):
                response = session.request(
                    method=method,
                    url=f"http://127.0.0.1:{port}{path}",
                    data=JSONSerializer().dumps(body) if body is not None and method != "get" else None,
                    headers=wrong_auth,
                )
                self.assertTrue(response.status_code == 401)

    def _test_authenticated_end_to_end(
        self,
        port: int,
        security_scheme: SecurityScheme,
        auth_headers: dict[str, str] = None,
        wrong_auth_headers: dict[str, str] = None,
    ):
        """Test end-to-end with authentication"""
        thing_id = f"test-sec-{uuid.uuid4().hex[0:8]}"
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(
            forked=True,
            port=port,
            config={"cors": True},
            security_schemes=[security_scheme],
        )
        self.wait_until_server_ready(port=port)
        self._test_handlers_end_to_end(port=port, thing_id=thing_id, headers=auth_headers)
        self._test_invalid_auth_end_to_end(port=port, thing_id=thing_id, wrong_auth_headers=wrong_auth_headers)
        # reinstate correct credentials to stop
        self.stop_server(port=port, thing_ids=[thing_id], headers=auth_headers)

    def test_06_basic_end_to_end(self):
        thing_id = f"test-sec-{uuid.uuid4().hex[0:8]}"
        port = 60004
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(forked=True, port=port, config={"cors": True})
        self.wait_until_server_ready(port=port)

        self._test_handlers_end_to_end(port=port, thing_id=thing_id, headers={"Content-Type": "application/json"})
        self.stop_server(port, thing_ids=[thing_id])

    def test_07_bcrypt_basic_security_end_to_end(self):
        security_scheme = BcryptBasicSecurity(username="someuser", password="somepassword")
        port = 60005
        self._test_authenticated_end_to_end(
            port=port,
            security_scheme=security_scheme,
            auth_headers={
                "Content-type": "application/json",
                "Authorization": f"Basic {base64.b64encode(b'someuser:somepassword').decode('utf-8')}",
            },
            wrong_auth_headers=[
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'wronguser:wrongpassword').decode('utf-8')}",
                },
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'someuser:wrongpassword').decode('utf-8')}",
                },
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'wronguser:somepassword').decode('utf-8')}",
                },
            ],
        )

    def test_08_argon2_basic_security_end_to_end(self):
        security_scheme = Argon2BasicSecurity(username="someuserargon2", password="somepasswordargon2")
        port = 60006
        self._test_authenticated_end_to_end(
            port=port,
            security_scheme=security_scheme,
            auth_headers={
                "Content-type": "application/json",
                "Authorization": f"Basic {base64.b64encode(b'someuserargon2:somepasswordargon2').decode('utf-8')}",
            },
            wrong_auth_headers=[
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'wronguserargon2:wrongpasswordargon2').decode('utf-8')}",
                },
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'someuserargon2:wrongpasswordargon2').decode('utf-8')}",
                },
                {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'wronguserargon2:somepasswordargon2').decode('utf-8')}",
                },
            ],
        )

    def _test_sse_end_to_end(
        self,
        port: int,
        security_scheme: SecurityScheme = None,
        headers: dict[str, str] = None,
    ):
        """
        Test end-to-end with Server-Sent Events (SSE).
        """
        thing_id = f"test-sse-{uuid.uuid4().hex[0:8]}"
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(
            forked=True,
            port=port,
            config={"cors": True},
            security_schemes=[security_scheme] if security_scheme else None,
        )
        self.wait_until_server_ready(port=port)

        session = requests.Session()
        response = session.post(f"http://127.0.0.1:{port}/{thing_id}/start-acquisition", headers=headers)
        self.assertEqual(response.status_code, 200)
        sse_gen = self.sse_stream(
            f"http://127.0.0.1:{port}/{thing_id}/intensity-measurement-event",
            headers=headers,
        )
        for i in range(5):
            evt = next(sse_gen)
            self.assertTrue("exception" not in evt)
        response = session.post(f"http://127.0.0.1:{port}/{thing_id}/stop-acquisition", headers=headers)
        self.stop_server(port=port, thing_ids=[thing_id], headers=headers)

    def test_09_sse(self):
        """Test Server-Sent Events (SSE)"""
        for security_scheme, port in [
            (None, 60007),
            (BcryptBasicSecurity(username="someuser", password="somepassword"), 60008),
        ]:
            # test SSE with and without security
            if security_scheme:
                headers = {
                    "Content-type": "application/json",
                    "Authorization": f"Basic {base64.b64encode(b'someuser:somepassword').decode('utf-8')}",
                }
            else:
                headers = dict()
            self._test_sse_end_to_end(port=port, security_scheme=security_scheme, headers=headers)

    def test_10_forms_generation(self):
        thing_id = f"test-forms-{uuid.uuid4().hex[0:8]}"
        port = 60009
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(forked=True, port=port, config={"cors": True})
        self.wait_until_server_ready(port=port)

        session = requests.Session()
        response = session.get(f"http://127.0.0.1:{port}/{thing_id}/resources/wot-td")
        self.assertEqual(response.status_code, 200)
        td = response.json()
        self.assertIn("properties", td)
        self.assertIn("actions", td)
        self.assertIn("events", td)
        self.assertTrue(len(td["properties"]) >= 0)
        self.assertTrue(len(td["actions"]) >= 0)
        self.assertTrue(len(td["events"]) >= 0)
        for prop in list(td["properties"].values()) + list(td["actions"].values()) + list(td["events"].values()):
            self.assertIn("forms", prop)
            self.assertTrue(len(prop["forms"]) > 0)
            for form in prop["forms"]:
                self.assertIn("href", form)
                self.assertIn("htv:methodName", form)
                self.assertIn("contentType", form)
                self.assertIn("op", form)
        self.stop_server(port=port, thing_ids=[thing_id])

    def test_11_object_proxy_basic(self):
        thing_id = f"test-obj-proxy-{uuid.uuid4().hex[0:8]}"
        port = 60010
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(forked=True, port=port, config={"cors": True})
        self.wait_until_server_ready(port=port)

        object_proxy = ClientFactory.http(url=f"http://127.0.0.1:{port}/{thing_id}/resources/wot-td")
        self.assertIsInstance(object_proxy, ObjectProxy)
        self.assertEqual(object_proxy.test_echo("Hello World!"), "Hello World!")
        self.assertEqual(
            asyncio.run(object_proxy.async_invoke_action("test_echo", "Hello World!")),
            "Hello World!",
        )
        self.assertEqual(object_proxy.read_property("max_intensity"), 16384)
        self.assertEqual(object_proxy.write_property("integration_time", 1200), None)
        self.assertEqual(object_proxy.read_property("integration_time"), 1200)
        self.stop_server(port=port, thing_ids=[thing_id])

    def notest_12_object_proxy_with_basic_auth(self):
        security_scheme = BcryptBasicSecurity(username="cliuser", password="clipass")
        port = 60013
        thing_id = f"test-basic-proxy-{uuid.uuid4().hex[0:8]}"
        thing = OceanOpticsSpectrometer(id=thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        thing.run_with_http_server(
            forked=True,
            port=port,
            config={"cors": True},
            security_schemes=[security_scheme],
        )
        self.wait_until_server_ready(port=port)

        object_proxy = ClientFactory.http(
            url=f"http://127.0.0.1:{port}/{thing_id}/resources/wot-td",
            username="cliuser",
            password="clipass",
        )
        self.assertEqual(object_proxy.read_property("max_intensity"), 16384)
        headers = {}
        token = base64.b64encode("cliuser:clipass".encode("utf-8")).decode("ascii")
        headers["Authorization"] = f"Basic {token}"
        self.stop_server(port=port, thing_ids=[thing_id], headers=headers)

    @classmethod
    def stop_server(cls, port, thing_ids: list[str] = [], **request_kwargs):
        session = requests.Session()
        endpoints = [("post", f"/{thing_id}/exit", None) for thing_id in thing_ids]
        endpoints += [("post", "/stop", None)]
        for method, path, body in endpoints:
            response = session.request(method=method, url=f"http://127.0.0.1:{port}{path}", **request_kwargs)
            if response.status_code not in [200, 201, 202, 204]:
                logging.warning(f"Failed to stop server or thing at {path} with status {response.status_code}")

    @classmethod
    def wait_until_server_ready(cls, port, tries: int = 10):
        session = requests.Session()
        for i in range(tries):
            try:
                response = session.get(f"http://127.0.0.1:{port}/liveness")
                if response.status_code in [200, 201, 202, 204]:
                    response = session.get(f"http://127.0.0.1:{port}/readiness")
                    if response.status_code in [200, 201, 202, 204]:
                        time.sleep(2)
                        return
            except Exception:
                pass
            time.sleep(1)
        raise TimeoutError(f"Server on port {port} not ready after {tries} tries")

    @classmethod
    def sse_stream(cls, url, chunk_size=2048, **kwargs):
        """Generator yielding dicts with the fields of each SSE event"""
        with requests.get(url, stream=True, **kwargs) as resp:
            resp.raise_for_status()
            buffer = ""
            for chunk in resp.iter_content(chunk_size=chunk_size, decode_unicode=True):
                buffer += chunk
                # split events on the SSE separator: two newlines
                while "\n\n" in buffer:
                    raw_event, buffer = buffer.split("\n\n", 1)
                    event = {}
                    for line in raw_event.splitlines():
                        # skip comments
                        if not line or line.startswith(":"):
                            continue
                        if ":" in line:
                            field, value = line.split(":", 1)
                            event.setdefault(field, "")
                            # strip leading space after colon
                            event[field] += value.lstrip()
                yield event

    @classmethod
    def generate_endpoints_for_thing(cls, class_: ThingMeta, thing_id: str) -> list[tuple[str, str, Any]]:
        if class_ == OceanOpticsSpectrometer:
            return [
                # read Property
                ("get", f"/{thing_id}/max-intensity", 16384),
                ("get", f"/{thing_id}/serial-number", "simulation"),
                # write Property
                ("put", f"/{thing_id}/integration-time", 1200),
                ("get", f"/{thing_id}/integration-time", 1200),
                # invoke action
                ("post", f"/{thing_id}/disconnect", None),
                ("post", f"/{thing_id}/connect", None),
            ]
        raise NotImplementedError(f"Endpoints for {class_.__name__} not implemented yet")


class TestHTTPObjectProxy(TestCase):
    # later create a TestObjtectProxy class that will test ObjectProxy but just overload the setUp and tearDown methods
    # with the different protocol

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.thing_id = f"test-obj-proxy-{uuid.uuid4().hex[0:8]}"
        cls.port = 60011
        cls.thing = OceanOpticsSpectrometer(id=cls.thing_id, serial_number="simulation", log_level=logging.ERROR + 10)
        cls.thing.run_with_http_server(forked=True, port=cls.port, config={"cors": True})
        TestHTTPServer.wait_until_server_ready(port=cls.port)

        cls.object_proxy = ClientFactory.http(url=f"http://127.0.0.1:{cls.port}/{cls.thing_id}/resources/wot-td")

    @classmethod
    def tearDownClass(cls):
        # stop the thing and server
        TestHTTPServer.stop_server(cls.port, thing_ids=[cls.thing.id])
        cls.object_proxy = None
        super().tearDownClass()

    def test_01_invoke_action(self):
        """Test basic functionality of ObjectProxy with HTTP server."""
        self.assertIsInstance(self.object_proxy, ObjectProxy)
        # Test invoke_action method with reply
        self.assertEqual(self.object_proxy.invoke_action("test_echo", "Hello World!"), "Hello World!")
        # Test invoke_action with dot notation
        self.assertEqual(self.object_proxy.test_echo(fake.chrome()), fake.last)
        self.assertEqual(self.object_proxy.test_echo(fake.sha256()), fake.last)
        self.assertEqual(self.object_proxy.test_echo(fake.address()), fake.last)
        # Test invoke_action with no reply
        self.assertEqual(
            self.object_proxy.invoke_action("test_echo", fake.random_number(), oneway=True),
            None,
        )
        # # Test invoke_action in non blocking mode
        noblock_payload = fake.pylist(20, value_types=[int, float, str, bool])
        noblock_msg_id = self.object_proxy.invoke_action("test_echo", noblock_payload, noblock=True)
        self.assertIsInstance(noblock_msg_id, str)
        self.assertEqual(
            self.object_proxy.invoke_action("test_echo", fake.pylist(20, value_types=[int, float, str, bool])),
            fake.last,
        )
        self.assertEqual(
            self.object_proxy.invoke_action("test_echo", fake.pylist(10, value_types=[int, float, str, bool])),
            fake.last,
        )
        self.assertEqual(self.object_proxy.read_reply(noblock_msg_id), noblock_payload)

    def test_02_rwd_properties(self):
        # test read and write properties
        self.assertEqual(self.object_proxy.read_property("max_intensity"), 16384)
        self.assertEqual(self.object_proxy.write_property("integration_time", 1200), None)
        self.assertEqual(self.object_proxy.read_property("integration_time"), 1200)
        # test read and write properties with dot notation
        self.assertEqual(self.object_proxy.max_intensity, 16384)
        self.assertEqual(self.object_proxy.integration_time, 1200)
        self.object_proxy.integration_time = 1000
        self.assertEqual(self.object_proxy.integration_time, 1000)
        # test oneway write property
        self.assertEqual(self.object_proxy.write_property("integration_time", 800, oneway=True), None)
        self.assertEqual(self.object_proxy.read_property("integration_time"), 800)
        # test noblock read property
        noblock_msg_id = self.object_proxy.read_property("integration_time", noblock=True)
        self.assertIsInstance(noblock_msg_id, str)
        self.assertEqual(self.object_proxy.read_property("max_intensity"), 16384)
        self.assertEqual(self.object_proxy.write_property("integration_time", 1200), None)
        self.assertEqual(self.object_proxy.read_reply(noblock_msg_id), 800)

    def notest_03_rw_multiple_properties(self):
        """Test reading and writing multiple properties at once."""
        # test read multiple properties
        properties = self.object_proxy.read_multiple_properties(["max_intensity", "integration_time"])
        self.assertEqual(properties["max_intensity"], 16384)
        self.assertEqual(properties["integration_time"], 800)

        # test write multiple properties
        new_values = {"integration_time": 1200, "max_intensity": 20000}
        self.object_proxy.write_multiple_properties(new_values)
        properties = self.object_proxy.read_multiple_properties(["max_intensity", "integration_time"])
        self.assertEqual(properties["max_intensity"], 20000)
        self.assertEqual(properties["integration_time"], 1200)

    def test_04_subscribe_event(self):
        """Test subscribing to an event and receiving updates."""
        event_name = "intensity_measurement_event"

        def on_event(data: SSE):
            nonlocal self
            self.assertTrue(isinstance(data.data, dict) and "value" in data.data and "timestamp" in data.data)

        self.object_proxy.subscribe_event(event_name, on_event)
        self.object_proxy.start_acquisition()
        time.sleep(2)  # wait for some events to be generated
        self.object_proxy.stop_acquisition()
        # check if events are kept alive
        time.sleep(20)
        self.object_proxy.start_acquisition()
        time.sleep(2)  # wait for some events to be generated
        self.object_proxy.stop_acquisition()
        self.object_proxy.unsubscribe_event(event_name)


class TestHTTPEndToEnd(TestRPCEndToEnd):
    @classmethod
    def setUpClass(cls):
        cls.http_port = 60012
        super().setUpClass()
        print("Test HTTP Object Proxy End to End")

    @classmethod
    def setUpThing(cls):
        """Set up the thing for the http object proxy client"""
        cls.thing = TestThing(id=cls.thing_id, log_level=logging.ERROR + 10)
        cls.thing.run_with_http_server(forked=True, port=cls.http_port, config={"cors": True})
        TestHTTPServer.wait_until_server_ready(port=cls.http_port)

        cls.thing_model = cls.thing.get_thing_model(ignore_errors=True).json()

    @classmethod
    def tearDownClass(cls):
        """Test the stop of the http object proxy client"""
        TestHTTPServer.stop_server(port=cls.http_port, thing_ids=[cls.thing_id])
        super().tearDownClass()

    @classmethod
    def get_client(cls):
        try:
            if cls._client is not None:
                return cls._client
            raise AttributeError()
        except AttributeError:
            cls._client = ClientFactory.http(
                url=f"http://127.0.0.1:{cls.http_port}/{cls.thing_id}/resources/wot-td", ignore_TD_errors=True
            )
            return cls._client

    def test_04_RW_multiple_properties(self):
        pass


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestHTTPServer))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestHTTPObjectProxy))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestHTTPEndToEnd))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))



import typing
import numpy
from types import FunctionType

from .parameterized import Parameterized, ParameterizedMetaclass
from .parameters import *


class NumpyArray(ClassSelector):
    """
    Parameter whose value is a numpy array.
    """

    def __init__(
        self,
        default=None,
        doc: typing.Union[str, None] = None,
        constant: bool = False,
        readonly: bool = False,
        allow_None: bool = False,
        label: typing.Union[str, None] = None,
        per_instance: bool = False,
        deep_copy: bool = False,
        class_member: bool = False,
        fget: FunctionType = None,
        fset: FunctionType = None,
        precedence: float = None,
    ) -> None:
        super(NumpyArray, self).__init__(
            class_=numpy.ndarray,
            default=default,
            doc=doc,
            constant=constant,
            readonly=readonly,
            allow_None=allow_None,
            label=label,
            per_instance=per_instance,
            deep_copy=deep_copy,
            class_member=class_member,
            fget=fget,
            fset=fset,
            precedence=precedence,
        )

    @typing.overload
    def __get__(
        self, obj: typing.Union[Parameterized, typing.Any], objtype: typing.Union[ParameterizedMetaclass, typing.Any]
    ) -> numpy.ndarray: ...

    @classmethod
    def serialize(cls, value: typing.Union[numpy.ndarray, None]):
        if value is None:
            return None
        return value.tolist()

    @classmethod
    def deserialize(cls, value):
        if value == "null" or value is None:
            return None
        return numpy.asarray(value)


from pandas import DataFrame as pdDFrame


class DataFrame(ClassSelector):
    """
    Parameter whose value is a pandas DataFrame.

    The structure of the DataFrame can be constrained by the rows and
    columns arguments:

    rows: If specified, may be a number or an integer bounds tuple to
    constrain the allowable number of rows.

    columns: If specified, may be a number, an integer bounds tuple, a
    list or a set. If the argument is numeric, constrains the number of
    columns using the same semantics as used for rows. If either a list
    or set of strings, the column names will be validated. If a set is
    used, the supplied DataFrame must contain the specified columns and
    if a list is given, the supplied DataFrame must contain exactly the
    same columns and in the same order and no other columns.
    """

    __slots__ = ["rows", "columns", "ordered"]

    def __init__(self, default=None, rows=None, columns=None, ordered=None, **params):
        self.rows = rows
        self.columns = columns
        self.ordered = ordered
        super(DataFrame, self).__init__(pdDFrame, default=default, **params)
        self._validate(self.default)

    def _length_bounds_check(self, bounds, length, name):
        message = "{name} length {length} does not match declared bounds of {bounds}"
        if not isinstance(bounds, tuple):
            if bounds != length:
                raise ValueError(message.format(name=name, length=length, bounds=bounds))
            else:
                return
        (lower, upper) = bounds
        failure = (lower is not None and (length < lower)) or (upper is not None and length > upper)
        if failure:
            raise ValueError(message.format(name=name, length=length, bounds=bounds))

    def _validate(self, val):
        super(DataFrame, self)._validate(val)

        if isinstance(self.columns, set) and self.ordered is True:
            raise ValueError("Columns cannot be ordered when specified as a set")

        if self.allow_None and val is None:
            return

        if self.columns is None:
            pass
        elif (
            isinstance(self.columns, tuple)
            and len(self.columns) == 2
            and all(isinstance(v, (type(None), numbers.Number)) for v in self.columns)
        ):  # Numeric bounds tuple
            self._length_bounds_check(self.columns, len(val.columns), "Columns")
        elif isinstance(self.columns, (list, set)):
            self.ordered = isinstance(self.columns, list) if self.ordered is None else self.ordered
            difference = set(self.columns) - set([str(el) for el in val.columns])
            if difference:
                msg = "Provided DataFrame columns {found} does not contain required columns {expected}"
                raise ValueError(msg.format(found=list(val.columns), expected=sorted(self.columns)))
        else:
            self._length_bounds_check(self.columns, len(val.columns), "Column")

        if self.ordered:
            if list(val.columns) != list(self.columns):
                msg = "Provided DataFrame columns {found} must exactly match {expected}"
                raise ValueError(msg.format(found=list(val.columns), expected=self.columns))

        if self.rows is not None:
            self._length_bounds_check(self.rows, len(val), "Row")

    @classmethod
    def serialize(cls, value):
        if value is None:
            return "null"
        return value.to_dict("records")

    @classmethod
    def deserialize(cls, value):
        if value == "null":
            return None
        from pandas import DataFrame as pdDFrame

        return pdDFrame(value)


class Series(ClassSelector):
    """
    Parameter whose value is a pandas Series.

    The structure of the Series can be constrained by the rows argument
    which may be a number or an integer bounds tuple to constrain the
    allowable number of rows.
    """

    __slots__ = ["rows"]

    def __init__(self, default=None, rows=None, allow_None=False, **params):
        from pandas import Series as pdSeries

        self.rows = rows
        super(Series, self).__init__(pdSeries, default=default, allow_None=allow_None, **params)
        self._validate(self.default)

    def _length_bounds_check(self, bounds, length, name):
        message = "{name} length {length} does not match declared bounds of {bounds}"
        if not isinstance(bounds, tuple):
            if bounds != length:
                raise ValueError(message.format(name=name, length=length, bounds=bounds))
            else:
                return
        (lower, upper) = bounds
        failure = (lower is not None and (length < lower)) or (upper is not None and length > upper)
        if failure:
            raise ValueError(message.format(name=name, length=length, bounds=bounds))

    def _validate(self, val):
        super(Series, self)._validate(val)

        if self.allow_None and val is None:
            return

        if self.rows is not None:
            self._length_bounds_check(self.rows, len(val), "Row")



"""
The following is a list of all dataclasses used to store information on the exposed
resources on the network. These classese are generally not for consumption by the package-end-user.
"""

import typing
from enum import Enum
from types import FunctionType, MethodType

from pydantic import BaseModel, RootModel

from ..param.parameters import String, Boolean, Tuple, ClassSelector, Parameter
from ..param.parameterized import ParameterizedMetaclass
from ..constants import USE_OBJECT_NAME
from ..schema_validators import BaseSchemaValidator
from ..utils import issubklass


class RemoteResourceInfoValidator:
    """
    A validator class for saving remote access related information on a resource. Currently callables (functions,
    methods and those with__call__) and class/instance property store this information as their own attribute under
    the variable ``_execution_info_validator``. This is later split into information suitable for HTTP server, ZMQ client & ``EventLoop``.

    Attributes
    ----------
    state : str, default None
        State machine state at which a callable will be executed or attribute/property can be
        written. Does not apply to read-only attributes/properties.
    obj_name : str, default - extracted object name
        the name of the object which will be supplied to the ``ObjectProxy`` class to populate
        its own namespace. For HTTP clients, HTTP method and URL path is important and for
        object proxies clients, the obj_name is important.
    isaction : bool, default False
        True for a method or function or callable
    isproperty : bool, default False
        True for a property
    """

    state = Tuple(
        default=None,
        item_type=(Enum, str),
        allow_None=True,
        accept_list=True,
        accept_item=True,
        doc="State machine state at which a callable will be executed or attribute/property can be written.",
    )  # type: typing.Tuple[typing.Union[Enum, str]]

    obj = ClassSelector(
        default=None,
        allow_None=True,
        class_=(
            FunctionType,
            MethodType,
            classmethod,
            Parameter,
            ParameterizedMetaclass,
        ),  # Property will need circular import so we stick to base class Parameter
        doc="the unbound object like the unbound method",
    )

    obj_name = String(
        default=USE_OBJECT_NAME,
        doc="the name of the object which will be supplied to the ``ObjectProxy`` class to populate its own namespace.",
    )  # type: str

    isaction = Boolean(default=False, doc="True for a method or function or callable")  # type: bool

    isproperty = Boolean(default=False, doc="True for a property")  # type: bool

    def __init__(self, **kwargs) -> None:
        # No full-scale checks for unknown keyword arguments as the class
        # is used by the developer, so please try to be error-proof
        for key, value in kwargs.items():
            setattr(self, key, value)


class ActionInfoValidator(RemoteResourceInfoValidator):
    """
    request_as_argument : bool, default False
        if True, http/ZMQ request object will be passed as an argument to the callable.
        The user is warned to not use this generally.
    argument_schema: JSON, default None
        JSON schema validations for arguments of a callable. Assumption is therefore arguments will be JSON complaint.
    return_value_schema: JSON, default None
        schema for return value of a callable. Assumption is therefore return value will be JSON complaint.
    create_task: bool, default True
        default for async methods/actions
    safe: bool, default True
        metadata information whether the action is safe to execute
    idempotent: bool, default False
        metadata information whether the action is idempotent
    synchronous: bool, default True
        metadata information whether the action is synchronous
    """

    request_as_argument = Boolean(
        default=False,
        doc="if True, http/RPC request object will be passed as an argument to the callable.",
    )  # type: bool

    argument_schema = Parameter(
        default=None,
        allow_None=True,
        # due to schema validation, this has to be a dict, and not a special dict like TypedDict
        doc="JSON schema validations for arguments of a callable",
    )

    return_value_schema = Parameter(
        default=None,
        allow_None=True,
        # due to schema validation, this has to be a dict, and not a special dict like TypedDict
        doc="schema for return value of a callable",
    )

    create_task = Boolean(default=True, doc="should a coroutine be tasked or run in the same loop?")  # type: bool

    iscoroutine = Boolean(
        default=False,  # not sure if isFuture or isCoroutine is correct, something to fix later
        doc="whether the callable should be awaited",
    )  # type: bool

    safe = Boolean(default=True, doc="metadata information whether the action is safe to execute")  # type: bool

    idempotent = Boolean(default=False, doc="metadata information whether the action is idempotent")  # type: bool

    synchronous = Boolean(default=True, doc="metadata information whether the action is synchronous")  # type: bool

    isparameterized = Boolean(default=False, doc="True for a parameterized function")  # type: bool

    isclassmethod = Boolean(default=False, doc="True for a classmethod")  # type: bool

    schema_validator = ClassSelector(
        default=None,
        allow_None=True,
        class_=BaseSchemaValidator,
        doc="schema validator for the callable if to be validated server side",
    )  # type: BaseSchemaValidator

    @argument_schema.getter
    def _get_argument_schema(self):
        return getattr(self, "_argument_schema", None)

    @argument_schema.setter
    def _set_argument_schema(self, value):
        value = self.action_payload_schema_validator(value)
        setattr(self, "_argument_schema", value)

    @return_value_schema.getter
    def _get_return_value_schema(self):
        return getattr(self, "_return_value_schema", None)

    @return_value_schema.setter
    def _set_return_value_schema(self, value):
        value = self.action_payload_schema_validator(value)
        setattr(self, "_return_value_schema", value)

    def action_payload_schema_validator(self, value: typing.Any) -> typing.Any:
        if value is None or isinstance(value, dict) or issubklass(value, (BaseModel, RootModel)):
            return value
        raise TypeError("Schema must be None, a dict, or a subclass of BaseModel or RootModel")



import typing
from ..constants import JSON


class JSONSchema:
    """type restrictor converting python types to JSON schema types"""

    _allowed_types = ("string", "number", "integer", "boolean", "object", "array", None)

    _replacements = {
        int: "integer",
        float: "number",
        str: "string",
        bool: "boolean",
        dict: "object",
        list: "array",
        tuple: "array",
        set: "array",
        type(None): "null",
        Exception: {
            "type": "object",
            "properties": {
                "message": {"type": "string"},
                "type": {"type": "string"},
                "traceback": {"type": "array", "items": {"type": "string"}},
                "notes": {"type": ["string", "null"]},
            },
            "required": ["message", "type", "traceback"],
        },
    }

    _schemas = {}

    @classmethod
    def is_allowed_type(cls, typ: typing.Any) -> bool:
        """check if a certain base type has a JSON schema base type"""
        if typ in JSONSchema._replacements.keys():
            return True
        return False

    @classmethod
    def has_additional_schema_definitions(cls, typ: typing.Any) -> bool:
        """Check, if in additional to the JSON schema base type, additional schema definitions exists."""
        if typ in JSONSchema._schemas.keys():
            return True
        return False

    @classmethod
    def get_base_type(cls, typ: typing.Any) -> str:
        if not JSONSchema.is_allowed_type(typ):
            raise TypeError(
                f"Object for wot-td has invalid type for JSON conversion. Given type - {type(typ)}. "
                + "Use JSONSchema.register_replacements on hololinked.wot.td.JSONSchema object to recognise the type."
            )
        return JSONSchema._replacements[typ]

    @classmethod
    def register_type_replacement(
        self, type: typing.Any, json_schema_base_type: str, schema: typing.Optional[JSON] = None
    ) -> None:
        """
        Specify a python type to map to a specific JSON type.

        For example:
        - `JSONSchema.register_type_replacement(MyCustomObject, 'object', schema=MyCustomObject.schema())`
        - `JSONSchema.register_type_replacement(IPAddress, 'string')`
        - `JSONSchema.register_type_replacement(MyByteArray, 'array', schema=dict(items=dict(type="integer", minimum=0, maximum=255)))`
        - `JSONSchema.register_type_replacement(Image, 'string', schema=dict(contentEncoding='base64'))`

        Parameters
        ----------
        type: typing.Any
            The Python type to register. The python type must be hashable (can be stored as a key in a dictionary).
        json_schema_base_type: str
            The base JSON schema type to map the Python type to. One of
            ('string', 'number', 'integer', 'boolean', 'object', 'array', 'null').
        schema: typing.Optional[JSON]
            An optional JSON schema to use for the type.
        """
        if json_schema_base_type in JSONSchema._allowed_types:
            JSONSchema._replacements[type] = json_schema_base_type
            if schema is not None:
                if json_schema_base_type not in ("array", "object"):
                    raise ValueError(
                        f"schemas support only for array and object JSON schema types, your specified type - {type}."
                    )
                JSONSchema._schemas[type] = schema
        else:
            raise TypeError(
                "json schema replacement type must be one of allowed type - 'string', 'object', 'array', 'string', "
                + f"'number', 'integer', 'boolean', 'null'. Given value {json_schema_base_type}"
            )

    @classmethod
    def get_additional_schema_definitions(cls, typ: typing.Any):
        """schema for array and objects only supported"""
        if not JSONSchema.has_additional_schema_definitions(typ):
            raise ValueError(f"Schema for {typ} not provided. register one with JSONSchema.register_type_replacement()")
        return JSONSchema._schemas[typ]



class BaseProtocolServer:
    def add_thing(self, thing):
        pass

    def remove_thing(self, thing):
        pass

    def add_property(self, thing, prop):
        pass

    def add_action(self, thing, action):
        pass

    def add_event(self, thing, event):
        pass

    def build_thing_description(self, thing):
        pass



class BreakInnerLoop(Exception):
    """
    raised to break all inner loop, either of Idler or of RunWith or WithoutStateMachine
    """

    pass


class BreakAllLoops(Exception):
    """
    raised to exit the event loop itself
    """

    pass


class BreakLoop(Exception):
    """
    raise and catch to exit a loop from within another function or method
    """

    pass


class BreakFlow(Exception):
    """
    raised to break the flow of the program
    """

    pass


class StateMachineError(Exception):
    """
    raise to show errors while calling actions or writing properties in wrong state
    """

    pass


class DatabaseError(Exception):
    """
    raise to show database related errors
    """


__all__ = ["BreakInnerLoop", "BreakAllLoops", "BreakLoop", "BreakFlow", "StateMachineError", "DatabaseError"]



import logging
import typing
import datetime
import threading
import asyncio
import time
from collections import deque

from ..utils import get_default_logger
from .events import Event
from .properties import List
from .properties import Integer, Number
from .thing import Thing as RemoteObject
from .actions import action as remote_method


log_message_schema = {
    "type": "object",
    "properties": {
        "level": {"type": "string", "description": "log level, one of DEBUG, INFO, WARN, ERROR, CRITICAL"},
        "timestamp": {"type": "string", "description": "timestamp of the log entry"},
        "thread_id": {"type": "integer", "description": "ID of the thread that generated the log entry"},
        "message": {"type": "string", "description": "log message"},
    },
    "required": ["level", "timestamp", "thread_id", "message"],
    "additionalProperties": False,
}


class RemoteAccessHandler(logging.Handler, RemoteObject):
    """
    Log handler with remote access attached to `Thing`'s logger, capable of streaming the log entries as events.
    Set `remote_accessible_logger` to True in the `Thing` to enable this handler.
    The schema of the pushed logs is an array of objects,
    where each object has the following JSON schema:

    ```json
    {
        "type": "object",
        "properties": {
            "level": {
                "type": "string",
                "description": "log level, one of DEBUG, INFO, WARN, ERROR, CRITICAL"
            },
            "timestamp": {
                "type": "string",
                "description": "timestamp of the log entry"
            },
            "thread_id": {
                "type": "integer",
                "description": "ID of the thread that generated the log entry"
            },
            "message": {
                "type": "string",
                "description": "log message"
            }
        },
        "required": ["level", "timestamp", "thread_id", "message"],
        "additionalProperties": false
    }
    ```
    """

    def __init__(self, id: str = "logger", maxlen: int = 500, stream_interval: float = 1.0, **kwargs) -> None:
        """
        Parameters
        ----------
        id: str, default 'logger'
            id of the object, generally only one instance per `Thing` necessary, therefore defaults to
            'logger'
        maxlen: int, default 500
            history of log entries to store in RAM
        stream_interval: float, default 1.0
            when streaming logs using log-events endpoint, this value is the stream interval.
        **kwargs:
            additional keyword arguments, currently supports:

            - `len_debug`: int
                length of debug logs, default maxlen/5
            - `len_info`: int
                length of info logs, default maxlen/5
            - `len_warn`: int
                length of warn logs, default maxlen/5
            - `len_error`: int
                length of error logs, default maxlen/5
            - `len_critical`: int
                length of critical logs, default maxlen/5
        """
        RemoteObject.__init__(self, id=id, **kwargs)
        logging.Handler.__init__(self)
        self.set_maxlen(maxlen, **kwargs)
        self.stream_interval = stream_interval
        self.diff_logs = []
        self._push_events = False
        self._events_thread = None

    log_events = Event(doc="stream logs", schema=log_message_schema)

    stream_interval = Number(
        default=1.0,
        bounds=(0.025, 60.0),
        crop_to_bounds=True,
        step=0.05,
        doc="interval at which logs should be published to a client.",
    )  # type: float

    def get_maxlen(self):
        return self._maxlen

    def set_maxlen(self, value: int, **kwargs) -> None:
        self._maxlen = value
        self._debug_logs = deque(maxlen=kwargs.pop("len_debug", int(value / 5)))
        self._info_logs = deque(maxlen=kwargs.pop("len_info", int(value / 5)))
        self._warn_logs = deque(maxlen=kwargs.pop("len_warn", int(value / 5)))
        self._error_logs = deque(maxlen=kwargs.pop("len_error", int(value / 5)))
        self._critical_logs = deque(maxlen=kwargs.pop("len_critical", int(value / 5)))
        self._execution_logs = deque(maxlen=value)

    maxlen = Integer(
        default=100,
        bounds=(1, None),
        crop_to_bounds=True,
        fget=get_maxlen,
        fset=set_maxlen,
        doc="length of execution log history to store",
    )  # type: int

    @remote_method(
        input_schema={
            "scheduling": {"type": "string", "enum": ["threaded", "async"]},
            "stream_interval": {"type": "number", "minimum": 0.025},
        },
    )
    def push_events(self, scheduling: str = "threaded", stream_interval: float = 1) -> None:
        """
        Push log events to a client.

        Parameters
        ----------
        scheduling: str
            'threaded' or 'async'. threaded starts a new thread, async schedules a task to the
            main event loop.
        stream_interval: float
            push interval in seconds.
        """
        self.stream_interval = stream_interval
        if scheduling == "async":
            asyncio.get_event_loop().call_soon(lambda: asyncio.create_task(self._async_push_diff_logs()))
        elif scheduling == "threading":
            if self._events_thread is not None:  # dont create again if one is already running
                self._events_thread = threading.Thread(target=self._push_diff_logs)
                self._events_thread.start()
        else:
            raise ValueError(f"scheduling can only be 'threaded' or 'async'. Given value {scheduling}")

    @remote_method()
    def stop_events(self) -> None:
        """
        stop pushing log events
        """
        self._push_events = False
        if self._events_thread:  # coroutine variant will resolve automatically
            self._events_thread.join()
            self._owner.logger.debug(f"joined log event source with thread-id {self._events_thread.ident}.")
            self._events_thread = None

    def emit(self, record: logging.LogRecord) -> None:
        # automatically called when a log entry is created
        # log_entry = self.format(record)
        self.format(record)
        info = {
            "level": record.levelname,
            "timestamp": datetime.datetime.fromtimestamp(record.created).strftime("%Y-%m-%dT%H:%M:%S.%f"),
            "thread_id": threading.get_ident(),
            "message": record.msg,
        }
        if record.levelno < logging.INFO:
            self._debug_logs.appendleft(info)
        elif record.levelno >= logging.INFO and record.levelno < logging.WARN:
            self._info_logs.appendleft(info)
        elif record.levelno >= logging.WARN and record.levelno < logging.ERROR:
            self._warn_logs.appendleft(info)
        elif record.levelno >= logging.ERROR and record.levelno < logging.CRITICAL:
            self._error_logs.appendleft(info)
        elif record.levelno >= logging.CRITICAL:
            self._critical_logs.appendleft(info)
        self._execution_logs.appendleft(info)

        if self._push_events:
            self.diff_logs.insert(0, info)

    def _push_diff_logs(self) -> None:
        self._push_events = True
        while self._push_events:
            time.sleep(self.stream_interval)
            if len(self.diff_logs) > 0:
                self.log_events.push(self.diff_logs)
                self.diff_logs.clear()
        # give time to collect final logs with certainty
        # self.owner.logger.info(f"ending log event source with thread-id {threading.get_ident()}.")

    async def _async_push_diff_logs(self) -> None:
        while self._push_events:
            await asyncio.sleep(self.stream_interval)
            # TODO: can use an async EventPublisher in the future
            if len(self.diff_logs) > 0:
                self.log_events.push(self.diff_logs)
                self.diff_logs.clear()
        # self.owner.logger.info("ending log events.")

    debug_logs = List(default=[], readonly=True, fget=lambda self: self._debug_logs, doc="logs at logging.DEBUG level")  # type: list[typing.Dict[str, typing.Any]]

    warn_logs = List(default=[], readonly=True, fget=lambda self: self._warn_logs, doc="logs at logging.WARN level")  # type: list[typing.Dict[str, typing.Any]]

    info_logs = List(default=[], readonly=True, fget=lambda self: self._info_logs, doc="logs at logging.INFO level")  # type: list[typing.Dict[str, typing.Any]]

    error_logs = List(default=[], readonly=True, fget=lambda self: self._error_logs, doc="logs at logging.ERROR level")  # type: list[typing.Dict[str, typing.Any]]

    critical_logs = List(
        default=[], readonly=True, fget=lambda self: self._critical_logs, doc="logs at logging.CRITICAL level"
    )  # type: list[typing.Dict[str, typing.Any]]

    execution_logs = List(
        default=[],
        readonly=True,
        fget=lambda self: self._execution_logs,
        doc="logs at all levels accumulated in order of collection/execution",
    )  # type: list[typing.Dict[str, typing.Any]]


def prepare_object_logger(instance: RemoteObject, log_level: int, log_file: str, remote_access: bool = False) -> None:
    """
    Setup logger for the object with default settings. If a logger is already present, it is not recreated.
    If remote access is present, it is not recreated. This is a single-shot method to be run at __init__.

    Parameters
    ----------
    log_level: int
        logging level.
    log_file: str
        log file path. A FileHandler is attached to the logger if this is not None.
    remote_access: bool
        if True, a RemoteAccessHandler is attached to the logger.
    """
    if instance.logger is None:
        instance.logger = get_default_logger(
            instance.id, logging.INFO if not log_level else log_level, None if not log_file else log_file
        )

    if remote_access and not any(isinstance(handler, RemoteAccessHandler) for handler in instance.logger.handlers):
        instance._remote_access_loghandler = RemoteAccessHandler(
            id="logger", maxlen=500, emit_interval=1, logger=instance.logger
        )
        # we set logger=instance.logger because so that we dont recreate one for remote access handler
        instance.logger.addHandler(instance._remote_access_loghandler)

    if not isinstance(instance, RemoteAccessHandler):
        for handler in instance.logger.handlers:
            # if remote access is True or not, if such a handler is found, make it a sub thing
            if isinstance(handler, RemoteAccessHandler):
                instance._remote_access_loghandler = handler


class LogHistoryHandler(logging.Handler):
    """
    Log history handler. Add and remove this handler at specific points to hold specific logs that are generated
    between those points. Currently used by execution context within `RPCServer` where one can fetch the
    execution logs that were collected during a specific operation.
    """

    def __init__(self, log_list: typing.Optional[typing.List] = None):
        """
        Parameters
        ----------

        log_list: list, optional
            Initial set of log entries to start with. Optional, defaults to empty list.
        """
        super().__init__()
        self.log_list: typing.List[typing.Dict] = [] if not log_list else log_list

    def emit(self, record: logging.LogRecord):
        # log_entry = self.format(record)
        self.format(record)
        self.log_list.insert(
            0,
            {
                "level": record.levelname,
                "timestamp": datetime.datetime.fromtimestamp(record.created).strftime("%Y-%m-%dT%H:%M:%S.%f"),
                "thread_id": threading.get_ident(),
                "message": record.msg,
            },
        )


__all__ = [LogHistoryHandler.__name__, RemoteAccessHandler.__name__]



import typing
import unittest
import logging

from hololinked.core.actions import BoundAction
from hololinked.core.events import EventDispatcher
from hololinked.core.zmq.brokers import EventPublisher
from hololinked.core import Thing, ThingMeta, Action, Event, Property
from hololinked.core.meta import (
    DescriptorRegistry,
    PropertiesRegistry,
    ActionsRegistry,
    EventsRegistry,
)
from hololinked.core.zmq.rpc_server import RPCServer, prepare_rpc_server
from hololinked.core.properties import Parameter
from hololinked.core.state_machine import BoundFSM
from hololinked.utils import get_default_logger
from hololinked.core.logger import RemoteAccessHandler

try:
    from .things import OceanOpticsSpectrometer
    from .utils import TestCase, TestRunner
except ImportError:
    from things import OceanOpticsSpectrometer
    from utils import TestCase, TestRunner


"""
The tests in this file are for the initialization of the Thing class and its subclasses.
1. Test Thing class 
2. Test Thing subclass
3. Test ThingMeta metaclass
4. Test ActionRegistry class
5. Test EventRegistry class
6. Test PropertiesRegistry class
"""


class TestThingInit(TestCase):
    """Test Thing class which is the bread and butter of this package."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test Thing instantiation with {cls.__name__}")
        cls.thing_cls = Thing
        # using a variable called thing_cls because same tests are repeated for different thing class

    """
    Test sequence is as follows:
    1. Test id requirements 
    2. Test logger setup
    3. Test state and state_machine setup
    4. Test composition of subthings
    5. Test servers init
    6. Test thing model generation
    """

    def test_1_id(self):
        """Test id property of Thing class"""
        # req. 1. instance name must be a string and cannot be changed after set
        thing = self.thing_cls(id="test_id", log_level=logging.WARN)
        self.assertEqual(thing.id, "test_id")
        with self.assertRaises(ValueError):
            thing.id = "new_instance"
        with self.assertRaises(NotImplementedError):
            del thing.id
        # req. 2. regex is r'[A-Za-z]+[A-Za-z_0-9\-\/]*', simple URI like
        valid_ids = ["test_id", "A123", "valid_id-123", "another/valid-id"]
        invalid_ids = ["123_invalid", "invalid id", "invalid@id", ""]
        for valid_id in valid_ids:
            thing.properties.descriptors["id"].validate_and_adapt(valid_id)
        for invalid_id in invalid_ids:
            with self.assertRaises(ValueError):
                thing.properties.descriptors["id"].validate_and_adapt(invalid_id)

    def test_2_logger(self):
        """Test logger setup"""
        # req. 1. logger must have remote access handler if remote_accessible_logger is True
        logger = get_default_logger("test_logger", log_level=logging.WARN)
        thing = self.thing_cls(
            id="test_remote_accessible_logger",
            logger=logger,
            remote_accessible_logger=True,
        )
        self.assertEqual(thing.logger, logger)
        self.assertTrue(any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers))
        # Therefore also check the false condition
        logger = get_default_logger("test_logger_2", log_level=logging.WARN)
        thing = self.thing_cls(
            id="test_logger_without_remote_access",
            logger=logger,
            remote_accessible_logger=False,
        )
        self.assertFalse(any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers))
        # NOTE - logger is modifiable after instantiation

        # req. 2. logger is created automatically if not provided
        thing = self.thing_cls(id="test_logger_auto_creation", log_level=logging.WARN)
        self.assertIsNotNone(thing.logger)
        self.assertFalse(any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers))
        self.assertNotEqual(thing.logger, logger)  # not the above logger that we used.
        # remote accessible only when we ask for it
        thing = self.thing_cls(
            id="test_logger_auto_creation_2",
            log_level=logging.WARN,
            remote_accessible_logger=True,
        )
        self.assertIsNotNone(thing.logger)
        self.assertTrue(any(isinstance(handler, RemoteAccessHandler) for handler in thing.logger.handlers))
        self.assertNotEqual(thing.logger, logger)

    def test_3_state(self):
        """Test state and state_machine setup"""
        # req. 1. state property must be None when no state machine is present
        thing1 = self.thing_cls(id="test_no_state_machine", log_level=logging.WARN)
        self.assertIsNone(thing1.state)
        self.assertIsNone(thing1.state_machine)
        # detailed checks in another file

    def test_4_subthings(self):
        """Test composition"""
        thing = self.thing_cls(id="test_subthings", log_level=logging.WARN, remote_accessible_logger=True)
        # req. 1. subthings must be a dictionary
        self.assertIsInstance(thing.sub_things, dict)
        self.assertEqual(len(thing.sub_things), 1)  # logger
        # req. 2. subthings are always recomputed when accessed (at least thats the way it is right now),
        # so we can add new subthings anytime
        thing.another_thing = OceanOpticsSpectrometer(id="another_thing", log_level=logging.WARN)
        self.assertIsInstance(thing.sub_things, dict)
        self.assertEqual(len(thing.sub_things), 2)
        # req. 3. subthings must be instances of Thing and have the parent as owner
        for name, subthing in thing.sub_things.items():
            self.assertTrue(thing in subthing._owners)
            self.assertIsInstance(subthing, Thing)
            # req. 4. name of subthing must match name of the attribute
            self.assertTrue(hasattr(thing, name))

    def test_5_servers_init(self):
        """Test if servers can be initialized/instantiated"""
        # req. 1. rpc_server and event_publisher must be None when not run()
        thing = self.thing_cls(id="test_servers_init", log_level=logging.ERROR)
        self.assertIsNone(thing.rpc_server)
        self.assertIsNone(thing.event_publisher)
        # req. 2. rpc_server and event_publisher must be instances of their respective classes when run()
        prepare_rpc_server(thing, "IPC")
        self.assertIsInstance(thing.rpc_server, RPCServer)
        self.assertIsInstance(thing.event_publisher, EventPublisher)
        # exit to quit nicely
        thing.rpc_server.exit()
        thing.event_publisher.exit()


class TestOceanOpticsSpectrometer(TestThingInit):
    """test Thing subclass example"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.thing_cls = OceanOpticsSpectrometer

    # check docs of the parent class for the test sequence

    def test_3_state(self):
        """Test state and state_machine setup"""
        thing1 = self.thing_cls(id="test_state_machine", log_level=logging.WARN)
        # req. 1. state and state machine must be present because we create this subclass with a state machine
        self.assertIsNotNone(thing1.state)
        self.assertIsInstance(thing1.state_machine, BoundFSM)
        # req. 2. state and state machine must be different for different instances
        thing2 = self.thing_cls(id="test_state_machine_2", log_level=logging.WARN)
        # first check if state machine exists
        self.assertIsNotNone(thing2.state)
        self.assertIsInstance(thing2.state_machine, BoundFSM)
        # then check if they are different
        self.assertNotEqual(thing1.state_machine, thing2.state_machine)
        # until state is set, initial state is equal
        self.assertEqual(thing1.state, thing2.state)
        self.assertEqual(thing1.state_machine.initial_state, thing2.state_machine.initial_state)
        # after state is set, they are different
        thing1.state_machine.set_state(thing1.states.ALARM)
        self.assertNotEqual(thing1.state, thing2.state)
        self.assertNotEqual(thing1.state_machine, thing2.state_machine)
        # initial state is still same
        self.assertEqual(thing1.state_machine.initial_state, thing2.state_machine.initial_state)


class TestMetaclass(TestCase):
    """Test ThingMeta metaclass which instantiates a Thing (sub-)class"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        print(f"test ThingMeta with {cls.__name__}")

    """
    Test sequence is as follows:
    1. Test metaclass of Thing class
    2. Test registry creation and access which is currently the main purpose of the metaclass
    """

    def test_1_metaclass(self):
        """test metaclass of Thing class"""
        # req. 1 metaclass must be ThingMeta of any Thing class
        self.assertEqual(Thing.__class__, ThingMeta)
        self.assertEqual(OceanOpticsSpectrometer.__class__, ThingMeta)
        self.assertEqual(Thing.__class__, OceanOpticsSpectrometer.__class__)

    def test_2_registry_creation(self):
        """test registry creation and access which is currently the main purpose of the metaclass"""
        # req. 1. registry attributes must be instances of their respective classes
        self.assertIsInstance(Thing.properties, PropertiesRegistry)
        self.assertIsInstance(Thing.actions, ActionsRegistry)
        self.assertIsInstance(Thing.events, EventsRegistry)

        # req. 2. new registries are not created on the fly and are same between accesses
        self.assertEqual(Thing.properties, Thing.properties)
        self.assertEqual(Thing.actions, Thing.actions)
        self.assertEqual(Thing.events, Thing.events)
        # This test is done as the implementation deviates from `param`

        # req. 3. different subclasses have different registries
        self.assertNotEqual(Thing.properties, OceanOpticsSpectrometer.properties)
        self.assertNotEqual(Thing.actions, OceanOpticsSpectrometer.actions)
        self.assertNotEqual(Thing.events, OceanOpticsSpectrometer.events)

        # create instances for further tests
        thing = Thing(id="test_registry_creation", log_level=logging.WARN)
        spectrometer = OceanOpticsSpectrometer(id="test_registry_creation_2", log_level=logging.WARN)

        # req. 4. registry attributes must be instances of their respective classes also for instances
        self.assertIsInstance(thing.properties, PropertiesRegistry)
        self.assertIsInstance(thing.actions, ActionsRegistry)
        self.assertIsInstance(thing.events, EventsRegistry)

        # req. 5. registries are not created on the fly and are same between accesses also for instances
        self.assertEqual(thing.properties, thing.properties)
        self.assertEqual(thing.actions, thing.actions)
        self.assertEqual(thing.events, thing.events)

        # req. 6. registries are not shared between instances
        self.assertNotEqual(thing.properties, spectrometer.properties)
        self.assertNotEqual(thing.actions, spectrometer.actions)
        self.assertNotEqual(thing.events, spectrometer.events)

        # req. 7. registries are not shared between instances and their classes
        self.assertNotEqual(thing.properties, Thing.properties)
        self.assertNotEqual(thing.actions, Thing.actions)
        self.assertNotEqual(thing.events, Thing.events)
        self.assertNotEqual(spectrometer.properties, OceanOpticsSpectrometer.properties)
        self.assertNotEqual(spectrometer.actions, OceanOpticsSpectrometer.actions)
        self.assertNotEqual(spectrometer.events, OceanOpticsSpectrometer.events)


# Uncomment the following for type hints while coding registry tests,
# comment it before testing, otherwise tests will fail due to overriding Thing object
# class Thing(Thing):
#     class_registry: PropertiesRegistry | ActionsRegistry | EventsRegistry
#     instance_registry: PropertiesRegistry  | ActionsRegistry | EventsRegistry  | None
#     descriptor_object: type[Property | Action | Event]

# class OceanOpticsSpectrometer(OceanOpticsSpectrometer):
#     class_registry: PropertiesRegistry | ActionsRegistry | EventsRegistry
#     instance_registry: PropertiesRegistry  | ActionsRegistry | EventsRegistry  | None
#     descriptor_object: type[Property | Action | Event]


class TestRegistry(TestCase):
    # Read the commented section above before proceeding to this test

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.setUpRegistryObjects()
        cls.setUpRegistryAttributes()
        if cls.is_abstract_test_class:
            return
        print(f"test {cls.registry_cls.__name__} with {cls.__name__}")

    @classmethod
    def setUpRegistryObjects(cls):
        cls.registry_cls = None  # type: DescriptorRegistry | None
        cls.registry_object = None  # type: type[Property | Action | Event]

    @property
    def is_abstract_test_class(self):
        # if self.registry_cls is None:
        #     print("registry_cls is None")
        # if self.registry_object is None:
        #     print("registry_object is None")
        return self.registry_cls is None or self.registry_object is None

    @classmethod
    def setUpRegistryAttributes(cls):
        if cls.registry_cls is None or cls.registry_object is None:
            return

        # create instances for further tests
        cls.thing = Thing(id=f"test_{cls.registry_object.__name__}_registry", log_level=logging.WARN)
        cls.spectrometer = OceanOpticsSpectrometer(
            id=f"test_{cls.registry_object.__name__}_registry", log_level=logging.WARN
        )
        if cls.registry_cls == ActionsRegistry:
            Thing.class_registry = Thing.actions
            OceanOpticsSpectrometer.class_registry = OceanOpticsSpectrometer.actions
            cls.thing.instance_registry = cls.thing.actions
            cls.spectrometer.instance_registry = cls.spectrometer.actions
            cls.bound_object = BoundAction
        elif cls.registry_cls == PropertiesRegistry:
            Thing.class_registry = Thing.properties
            OceanOpticsSpectrometer.class_registry = OceanOpticsSpectrometer.properties
            cls.thing.instance_registry = cls.thing.properties
            cls.spectrometer.instance_registry = cls.spectrometer.properties
            cls.bound_object = typing.Any
        elif cls.registry_cls == EventsRegistry:
            Thing.class_registry = Thing.events
            OceanOpticsSpectrometer.class_registry = OceanOpticsSpectrometer.events
            cls.thing.instance_registry = cls.thing.events
            cls.spectrometer.instance_registry = cls.spectrometer.events
            cls.bound_object = EventDispatcher
        else:
            raise NotImplementedError("This registry class is not implemented")

    """
    Test action registry first because actions are the easiest to test.
    1. Test owner attribute
    2. Test descriptors access 
    3. Test dunders
    """

    def test_1_owner(self):
        """Test owner attribute of DescriptorRegistry"""
        if self.is_abstract_test_class:
            return
        # See comment above TestRegistry class to enable type definitions
        # req. 1. owner attribute must be the class itself when accessed as class attribute
        self.assertEqual(Thing.class_registry.owner, Thing)
        self.assertEqual(OceanOpticsSpectrometer.class_registry.owner, OceanOpticsSpectrometer)
        # therefore owner instance must be None
        self.assertIsNone(Thing.class_registry.owner_inst)
        self.assertIsNone(OceanOpticsSpectrometer.class_registry.owner_inst)

        # req. 2. owner attribute must be the instance for instance registries (i.e. when accessed as instance attribute)
        self.assertEqual(self.thing.instance_registry.owner, self.thing)
        self.assertEqual(self.spectrometer.instance_registry.owner, self.spectrometer)
        self.assertEqual(self.thing.instance_registry.owner_cls, Thing)
        self.assertEqual(self.spectrometer.instance_registry.owner_cls, OceanOpticsSpectrometer)

        # req. 3. descriptor_object must be defined correctly and is a class
        self.assertEqual(Thing.class_registry.descriptor_object, self.registry_object)
        self.assertEqual(
            OceanOpticsSpectrometer.class_registry.descriptor_object,
            self.registry_object,
        )
        self.assertEqual(self.thing.instance_registry.descriptor_object, self.registry_object)
        self.assertEqual(
            self.thing.instance_registry.descriptor_object,
            Thing.class_registry.descriptor_object,
        )

    def test_2_descriptors(self):
        """Test descriptors access"""
        if self.is_abstract_test_class:
            return

        # req. 1. descriptors are instances of the descriptor object - Property | Action | Event
        for name, value in Thing.class_registry.descriptors.items():
            self.assertIsInstance(value, self.registry_object)
            self.assertIsInstance(name, str)
        for name, value in OceanOpticsSpectrometer.class_registry.descriptors.items():
            self.assertIsInstance(value, self.registry_object)
            self.assertIsInstance(name, str)
        # subclass have more descriptors than parent class because our example Thing OceanOpticsSpectrometer
        # has defined its own actions, properties and events
        self.assertTrue(len(OceanOpticsSpectrometer.class_registry.descriptors) > len(Thing.class_registry.descriptors))
        # req. 2. either class level or instance level descriptors are same - not a strict requirement for different
        # use cases, one can always add instance level descriptors
        for name, value in self.thing.instance_registry.descriptors.items():
            self.assertIsInstance(value, self.registry_object)
            self.assertIsInstance(name, str)
        for name, value in self.spectrometer.instance_registry.descriptors.items():
            self.assertIsInstance(value, self.registry_object)
            self.assertIsInstance(name, str)
        # req. 3. because class level and instance level descriptors are same, they are equal
        for (name, value), (name2, value2) in zip(
            Thing.class_registry.descriptors.items(),
            self.thing.instance_registry.descriptors.items(),
        ):
            self.assertEqual(name, name2)
            self.assertEqual(value, value2)
        for (name, value), (name2, value2) in zip(
            OceanOpticsSpectrometer.class_registry.descriptors.items(),
            self.spectrometer.instance_registry.descriptors.items(),
        ):
            self.assertEqual(name, name2)
            self.assertEqual(value, value2)
        # req. 4. descriptors can be cleared
        self.assertTrue(
            hasattr(
                self.thing.instance_registry,
                f"_{self.thing.instance_registry._qualified_prefix}_{self.registry_cls.__name__.lower()}",
            )
        )
        self.thing.instance_registry.clear()
        self.assertTrue(
            not hasattr(
                self.thing.instance_registry,
                f"_{self.thing.instance_registry._qualified_prefix}_{self.registry_cls.__name__.lower()}",
            )
        )
        # clearing again any number of times should not raise error
        self.thing.instance_registry.clear()
        self.thing.instance_registry.clear()
        self.assertTrue(
            not hasattr(
                self.thing.instance_registry,
                f"_{self.thing.instance_registry._qualified_prefix}_{self.registry_cls.__name__.lower()}",
            )
        )

    def test_3_dunders(self):
        """Test dunders of DescriptorRegistry"""
        if self.is_abstract_test_class:
            return

        # req. 1. __getitem__ must return the descriptor object
        for name, value in Thing.class_registry.descriptors.items():
            self.assertEqual(Thing.class_registry[name], value)
            # req. 2. __contains__ must return True if the descriptor is present
            self.assertIn(value, Thing.class_registry)
            self.assertIn(name, Thing.class_registry.descriptors.keys())

        # req. 2. __iter__ must return an iterator over the descriptors dictionary
        # which in turn iterates over the keys
        self.assertTrue(all(isinstance(descriptor_name, str) for descriptor_name in Thing.class_registry))
        self.assertTrue(
            all(isinstance(descriptor_name, str) for descriptor_name in OceanOpticsSpectrometer.class_registry)
        )
        # __iter__ can also be casted as other iterators like lists
        thing_descriptors = list(self.thing.instance_registry)
        spectrometer_descriptors = list(self.spectrometer.instance_registry)
        self.assertIsInstance(thing_descriptors, list)
        self.assertIsInstance(spectrometer_descriptors, list)
        self.assertTrue(all(isinstance(descriptor_name, str) for descriptor_name in thing_descriptors))
        self.assertTrue(all(isinstance(descriptor_name, str) for descriptor_name in spectrometer_descriptors))

        # req. 3. __len__ must return the number of descriptors
        self.assertTrue(len(Thing.class_registry) == len(Thing.class_registry.descriptors))
        self.assertTrue(
            len(OceanOpticsSpectrometer.class_registry) == len(OceanOpticsSpectrometer.class_registry.descriptors)
        )
        self.assertTrue(len(self.thing.instance_registry) == len(self.thing.instance_registry.descriptors))
        self.assertTrue(
            len(self.spectrometer.instance_registry) == len(self.spectrometer.instance_registry.descriptors)
        )
        self.assertTrue(len(self.thing.instance_registry) == len(Thing.class_registry))
        self.assertTrue(len(self.spectrometer.instance_registry) == len(OceanOpticsSpectrometer.class_registry))

        # req. 4. registries have their unique hashes
        # NOTE - not sure if this is really a useful feature or just plain stupid
        # The requirement was to be able to generate unique hashes for each registry like foodict[<some hash>] = Thing.actions
        foodict = {
            Thing.class_registry: 1,
            OceanOpticsSpectrometer.class_registry: 2,
            self.thing.instance_registry: 3,
            self.spectrometer.instance_registry: 4,
        }
        self.assertEqual(foodict[Thing.class_registry], 1)
        self.assertEqual(foodict[OceanOpticsSpectrometer.class_registry], 2)
        self.assertEqual(foodict[self.thing.instance_registry], 3)
        self.assertEqual(foodict[self.spectrometer.instance_registry], 4)

        # __dir__ not yet tested
        # __str__ will not be tested

    def test_4_bound_objects(self):
        """Test bound objects returned from descriptor access"""
        if self.is_abstract_test_class:
            return
        if self.registry_object not in [Property, Parameter, Action]:
            # Events work a little differently, may need to be tested separately or refactored to same implementation
            return

        # req. 1. number of bound objects must be equal to number of descriptors
        # for example, number of bound actions must be equal to number of actions
        self.assertEqual(
            len(self.thing.instance_registry),
            len(self.thing.instance_registry.descriptors),
        )
        self.assertEqual(
            len(self.spectrometer.instance_registry),
            len(self.spectrometer.instance_registry.descriptors),
        )

        # req. 2. bound objects must be instances of bound instances
        for name, value in self.thing.instance_registry.values.items():
            if self.bound_object != typing.Any:
                self.assertIsInstance(value, self.bound_object)
            self.assertIsInstance(name, str)
        for name, value in self.spectrometer.instance_registry.values.items():
            if self.bound_object != typing.Any:
                self.assertIsInstance(value, self.bound_object)
            self.assertIsInstance(name, str)


class TestActionRegistry(TestRegistry):
    """Test ActionRegistry class"""

    @classmethod
    def setUpRegistryObjects(cls):
        cls.registry_cls = ActionsRegistry
        cls.registry_object = Action


class TestEventRegistry(TestRegistry):
    @classmethod
    def setUpRegistryObjects(cls):
        cls.registry_cls = EventsRegistry
        cls.registry_object = Event

    def test_2_descriptors(self):
        if self.is_abstract_test_class:
            return

        super().test_2_descriptors()

        # req. 5. observables and change events are also descriptors
        for name, value in self.thing.events.observables.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
        for name, value in self.thing.events.change_events.items():
            self.assertIsInstance(value, Event)
            self.assertIsInstance(name, str)
        # req. 4. descriptors can be cleared
        self.assertTrue(
            hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}",
            )
        )
        self.assertTrue(
            hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
            )
        )
        self.assertTrue(
            hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
            )
        )
        self.thing.events.clear()
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}",
            )
        )
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
            )
        )
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
            )
        )
        self.thing.events.clear()
        self.thing.events.clear()
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}",
            )
        )
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_change_events",
            )
        )
        self.assertTrue(
            not hasattr(
                self.thing.events,
                f"_{self.thing.events._qualified_prefix}_{EventsRegistry.__name__.lower()}_observables",
            )
        )


class TestPropertiesRegistry(TestRegistry):
    @classmethod
    def setUpRegistryObjects(cls):
        cls.registry_cls = PropertiesRegistry
        cls.registry_object = Parameter

    def test_2_descriptors(self):
        if self.is_abstract_test_class:
            return

        super().test_2_descriptors()

        # req. 5. parameters that are subclass of Property are usually remote objects
        for name, value in self.thing.properties.remote_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
        for name, value in self.spectrometer.properties.remote_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
        # req. 6. db_objects, db_init_objects, db_persisting_objects, db_commit_objects are also descriptors
        for name, value in self.thing.properties.db_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
            self.assertTrue(value.db_init or value.db_persist or value.db_commit)
        for name, value in self.thing.properties.db_init_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
            self.assertTrue(value.db_init or value.db_persist)
            self.assertFalse(value.db_commit)
        for name, value in self.thing.properties.db_commit_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
            self.assertTrue(value.db_commit or value.db_persist)
            self.assertFalse(value.db_init)
        for name, value in self.thing.properties.db_persisting_objects.items():
            self.assertIsInstance(value, Property)
            self.assertIsInstance(name, str)
            self.assertTrue(value.db_persist)
            self.assertFalse(value.db_init)  # in user given cases, this could be true, this is not strict requirement
            self.assertFalse(value.db_commit)  # in user given cases, this could be true, this is not strict requirement

        # req. 4. descriptors can be cleared
        self.assertTrue(
            hasattr(
                self.thing.properties,
                f"_{self.thing.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
            )
        )
        self.thing.properties.clear()
        self.assertTrue(
            not hasattr(
                self.thing.properties,
                f"_{self.thing.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
            )
        )
        self.thing.properties.clear()
        self.thing.properties.clear()
        self.assertTrue(
            not hasattr(
                self.thing.properties,
                f"_{self.thing.properties._qualified_prefix}_{PropertiesRegistry.__name__.lower()}",
            )
        )

    def test_5_bulk_read_write(self):
        """Test bulk read and write operations for properties"""

        # req. 1. test read in bulk for readAllProperties
        prop_values = self.spectrometer.properties.get()
        # read value is a dictionary
        self.assertIsInstance(prop_values, dict)
        self.assertTrue(len(prop_values) > 0)
        # all properties are read at instance level and get only reads remote objects
        self.assertTrue(len(prop_values) == len(self.spectrometer.properties.remote_objects))
        # read values are not descriptors themselves
        for name, value in prop_values.items():
            self.assertIsInstance(name, str)
            self.assertNotIsInstance(value, Parameter)  # descriptor has been read

        # req. 2. properties can be read with new names
        prop_values = self.spectrometer.properties.get(
            integration_time="integrationTime",
            state="State",
            trigger_mode="triggerMode",
        )
        self.assertIsInstance(prop_values, dict)
        self.assertTrue(len(prop_values) == 3)
        for name, value in prop_values.items():
            self.assertIsInstance(name, str)
            self.assertTrue(name in ["integrationTime", "triggerMode", "State"])
            self.assertNotIsInstance(value, Parameter)

        # req. 3. read in bulk for readMultipleProperties
        prop_values = self.spectrometer.properties.get(
            names=["integration_time", "trigger_mode", "state", "last_intensity"]
        )
        # read value is a dictionary
        self.assertIsInstance(prop_values, dict)
        self.assertTrue(len(prop_values) == 4)
        # read values are not descriptors themselves
        for name, value in prop_values.items():
            self.assertIsInstance(name, str)
            self.assertTrue(name in ["integration_time", "trigger_mode", "state", "last_intensity"])
            self.assertNotIsInstance(value, Parameter)

        # req. 4. read a property that is not present raises AttributeError
        with self.assertRaises(AttributeError) as ex:
            prop_values = self.spectrometer.properties.get(
                names=[
                    "integration_time",
                    "trigger_mode",
                    "non_existent_property",
                    "last_intensity",
                ]
            )
        self.assertTrue("property non_existent_property does not exist" in str(ex.exception))

        # req. 5. write in bulk
        prop_values = self.spectrometer.properties.get()
        self.spectrometer.properties.set(integration_time=10, trigger_mode=1)
        self.assertNotEqual(prop_values["integration_time"], self.spectrometer.integration_time)
        self.assertNotEqual(prop_values["trigger_mode"], self.spectrometer.trigger_mode)

        # req. 6. writing a non existent property raises RuntimeError
        with self.assertRaises(RuntimeError) as ex:
            self.spectrometer.properties.set(integration_time=120, trigger_mode=2, non_existent_property=10)
        self.assertTrue("Some properties could not be set due to errors" in str(ex.exception))
        self.assertTrue("non_existent_property" in str(ex.exception.__notes__))
        # but those that exist will still be written
        self.assertEqual(self.spectrometer.integration_time, 120)
        self.assertEqual(self.spectrometer.trigger_mode, 2)

    def test_6_db_properties(self):
        """Test db operations for properties"""

        # req. 1. db operations are supported only at instance level
        with self.assertRaises(AttributeError) as ex:
            Thing.properties.load_from_DB()
        self.assertTrue("database operations are only supported at instance level" in str(ex.exception))
        with self.assertRaises(AttributeError) as ex:
            Thing.properties.get_from_DB()
        self.assertTrue("database operations are only supported at instance level" in str(ex.exception))


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestThingInit))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestOceanOpticsSpectrometer))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestMetaclass))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestActionRegistry))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestPropertiesRegistry))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestEventRegistry))
    return suite


if __name__ == "__main__":
    runner = TestRunner()
    runner.run(load_tests(unittest.TestLoader(), None, None))


"""
# Summary of tests and requirements:

TestThing class:
1. Test id requirements:
    - Instance name must be a string and cannot be changed after set.
    - Valid and invalid IDs based on regex (r'[A-Za-z]+[A-Za-z_0-9\\-\\/]*').
2. Test logger setup:
    - Logger must have remote access handler if remote_accessible_logger is True.
    - Logger is created automatically if not provided.
3. Test state and state_machine setup:
    - State property must be None when no state machine is present.
4. Test composition of subthings:
    - Subthings must be a dictionary.
    - Subthings are recomputed when accessed.
    - Subthings must be instances of Thing and have the parent as owner.
    - Name of subthing must match name of the attribute.
5. Test servers init:
    - rpc_server and event_publisher must be None when not run().
    - rpc_server and event_publisher must be instances of their respective classes when run().
6. Test thing model generation:
    - Basic test to ensure nothing is fundamentally wrong.

TestOceanOpticsSpectrometer class:
1. Test state and state_machine setup:
    - State and state machine must be present because subclass has a state machine.
    - State and state machine must be different for different instances.

TestMetaclass class:
1. Test metaclass of Thing class:
    - Metaclass must be ThingMeta for any Thing class.
2. Test registry creation and access:
    - Registry attributes must be instances of their respective classes.
    - New registries are not created on the fly and are same between accesses.
    - Different subclasses have different registries.
    - Registry attributes must be instances of their respective classes also for instances.
    - Registries are not created on the fly and are same between accesses also for instances.
    - Registries are not shared between instances.
    - Registries are not shared between instances and their classes.

TestRegistry class:
1. Test owner attribute:
    - Owner attribute must be the class itself when accessed as class attribute.
    - Owner attribute must be the instance for instance registries.
    - Descriptor_object must be defined correctly and is a class.
2. Test descriptors access:
    - Descriptors are instances of the descriptor object.
    - Class level or instance level descriptors are same.
    - Descriptors can be cleared.
3. Test dunders:
    - __getitem__ must return the descriptor object.
    - __contains__ must return True if the descriptor is present.
    - __iter__ must return an iterator over the descriptors dictionary.
    - __len__ must return the number of descriptors.
    - Registries have their unique hashes.
4. Test bound objects:
    - Number of bound objects must be equal to number of descriptors.
    - Bound objects must be instances of bound instances.

TestActionRegistry class:
- Inherits tests from TestRegistry.

TestEventRegistry class:
- Inherits tests from TestRegistry.
- Observables and change events are also descriptors.

TestPropertiesRegistry class:
- Inherits tests from TestRegistry.
- Parameters that are subclass of Property are usually remote objects.
- DB operations are supported only at instance level.
"""



import sys
import logging
import re
import asyncio
import inspect
import typing
import types
import traceback
import ifaddr
import threading
from functools import wraps
from collections import OrderedDict
from dataclasses import asdict
from pydantic import BaseModel, ConfigDict, create_model, Field, RootModel
from inspect import Parameter, signature


def get_IP_from_interface(interface_name: str = "Ethernet", adapter_name=None) -> str:
    """
    Get IP address of specified interface. Generally necessary when connected to the network
    through multiple adapters and a server binds to only one adapter at a time.

    Parameters
    ----------
    interface_name: str
        Ethernet, Wifi etc.
    adapter_name: optional, str
        name of the adapter if available

    Returns
    -------
    str:
        IP address of the interface
    """
    adapters = ifaddr.get_adapters(include_unconfigured=True)
    for adapter in adapters:
        if not adapter_name:
            for ip in adapter.ips:
                if interface_name == ip.nice_name:
                    if ip.is_IPv4:
                        return ip.ip
        elif adapter_name == adapter.nice_name:
            for ip in adapter.ips:
                if interface_name == ip.nice_name:
                    if ip.is_IPv4:
                        return ip.ip
    raise ValueError(f"interface name {interface_name} not found in system interfaces.")


def format_exception_as_json(exc: Exception) -> typing.Dict[str, typing.Any]:
    """
    return exception as a JSON serializable dictionary
    """
    return {
        "message": str(exc),
        "type": repr(exc).split("(", 1)[0],
        "traceback": traceback.format_exc().splitlines(),
        "notes": exc.__notes__ if hasattr(exc, "__notes__") else None,
    }


def pep8_to_dashed_name(word: str) -> str:
    """
    Make an underscored, lowercase form from the expression in the string.
    Example::
        >>> pep8_to_dashed_URL("device_type")
        'device-type'
    """
    val = re.sub(r"_+", "-", word.lstrip("_").rstrip("_"))
    return val.replace(" ", "-")


def get_default_logger(
    name: str,
    log_level: int = logging.INFO,
    log_file: str = None,
    format: str = "%(levelname)-8s - %(asctime)s:%(msecs)03d - %(name)s - %(message)s",
) -> logging.Logger:
    """
    the default logger used by most of hololinked package, when arguments are not modified.
    StreamHandler is always created, pass log_file for a FileHandler as well.

    Parameters
    ----------
    name: str
        name of logger
    log_level: int
        log level
    log_file: str
        valid path to file
    format: str
        log format

    Returns
    -------
    logging.Logger:
        created logger
    """
    from .config import global_config

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG if global_config.DEBUG else log_level)
    if not any(isinstance(handler, logging.StreamHandler) for handler in logger.handlers):
        default_handler = logging.StreamHandler(sys.stdout)
        default_handler.setFormatter(logging.Formatter(format, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(default_handler)
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(logging.Formatter(format, datefmt="%Y-%m-%dT%H:%M:%S"))
        logger.addHandler(file_handler)
    return logger


def run_coro_sync(coro: typing.Coroutine):
    """
    run coroutine synchronously
    """
    try:
        eventloop = asyncio.get_event_loop()
    except RuntimeError:
        eventloop = asyncio.new_event_loop()
        asyncio.set_event_loop(eventloop)
    if eventloop.is_running():
        raise RuntimeError(
            f"asyncio event loop is already running, cannot setup coroutine {coro.__name__} to run sync, please await it."
        )
        # not the same as RuntimeError catch above.
    else:
        return eventloop.run_until_complete(coro)


def run_callable_somehow(
    method: typing.Union[typing.Callable, typing.Coroutine],
) -> typing.Any:
    """
    run method if synchronous, or when async, either schedule a coroutine or run it until its complete
    """
    if inspect.isawaitable(method):
        coro = method  # already a coroutine/awaitable object
    elif callable(method):
        result = method()  # call it
        if inspect.isawaitable(result):
            coro = result
        else:
            return result  # truly synchronous
    else:
        raise TypeError("method must be a callable or an awaitable")
    eventloop = get_current_async_loop()
    if eventloop.is_running():
        # task =  # check later if lambda is necessary
        eventloop.create_task(coro)
    else:
        # task = method
        return eventloop.run_until_complete(coro)


def complete_pending_tasks_in_current_loop():
    """
    Complete all pending tasks in the current asyncio event loop.
    """
    get_current_async_loop().run_until_complete(asyncio.gather(*asyncio.all_tasks(get_current_async_loop())))


async def complete_pending_tasks_in_current_loop_async():
    """
    Complete all pending tasks in the current asyncio event loop.
    """
    await asyncio.gather(*asyncio.all_tasks(get_current_async_loop()))


def print_pending_tasks_in_current_loop():
    """
    Print all pending tasks in the current asyncio event loop.
    """
    tasks = asyncio.all_tasks(get_current_async_loop())
    if not tasks:
        print("No pending tasks in the current event loop.")
        return
    for task in tasks:
        print(f"Task: {task}, Status: {task._state}")


def get_signature(
    callable: typing.Callable,
) -> typing.Tuple[typing.List[str], typing.List[type]]:
    """
    Retrieve the names and types of arguments based on annotations for the given callable.

    Parameters
    ----------
    callable: Callable
        function or method (not tested with __call__)

    Returns
    -------
    tuple: List[str], List[type]
        arguments name and types respectively
    """
    arg_names = []
    arg_types = []

    for param in inspect.signature(callable).parameters.values():
        arg_name = param.name
        arg_type = param.annotation if param.annotation != inspect.Parameter.empty else None

        arg_names.append(arg_name)
        arg_types.append(arg_type)

    return arg_names, arg_types


def getattr_without_descriptor_read(instance, key):
    """
    supply to inspect._get_members (not inspect.get_members) to avoid calling
    __get__ on hardware attributes
    """
    if key in instance.__dict__:
        return instance.__dict__[key]
    mro = mro = (instance.__class__,) + inspect.getmro(instance.__class__)
    for base in mro:
        if key in base.__dict__:
            value = base.__dict__[key]
            if isinstance(value, types.FunctionType):
                method = getattr(instance, key, None)
                if isinstance(method, types.MethodType):
                    return method
            return value
    # for descriptor, first try to find it in class dict or instance dict (for instance descriptors (per_instance_descriptor=True))
    # and then getattr from the instance. For descriptors/property, it will be mostly at above two levels.
    return getattr(instance, key, None)  # we can deal with None where we use this getter, so dont raise AttributeError


def isclassmethod(method) -> bool:
    """
    Returns `True` if the method is a classmethod, `False` otherwise.
    https://stackoverflow.com/questions/19227724/check-if-a-function-uses-classmethod

    """
    if isinstance(method, classmethod):
        return True
    bound_to = getattr(method, "__self__", None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = method.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False


def has_async_def(method) -> bool:
    """
    Checks if async def is found in method signature. Especially useful for class methods.
    https://github.com/python/cpython/issues/100224#issuecomment-2000895467

    Parameters
    ----------
    method: Callable
        function or method

    Returns
    -------
    bool
        True if async def is found in method signature, False otherwise
    """
    source = inspect.getsource(method)
    if re.search(
        r"^\s*async\s+def\s+" + re.escape(method.__name__) + r"\s*\(",
        source,
        re.MULTILINE,
    ):
        return True
    return False


def issubklass(obj, cls) -> bool:
    """
    Safely check if `obj` is a subclass of `cls`.

    Parameters
    ----------
    obj: typing.Any
        The object to check if it's a subclass.
    cls: typing.Any
        The class (or tuple of classes) to compare against.

    Returns
    -------
    bool
        True if `obj` is a subclass of `cls`, False otherwise.
    """
    try:
        # Check if obj is a class or a tuple of classes
        if isinstance(obj, type):
            return issubclass(obj, cls)
        elif isinstance(obj, tuple):
            # Ensure all elements in the tuple are classes
            return all(isinstance(o, type) for o in obj) and issubclass(obj, cls)
        else:
            return False
    except TypeError:
        return False


def get_a_filename_from_instance(thing: type, extension: str = "json") -> str:
    class_name = thing.__class__.__name__

    # Remove invalid characters from the instance name
    safe_id = re.sub(r'[<>:"/\\|?*\x00-\x1F]+', "_", thing.id)
    # Collapse consecutive underscores into one
    safe_id = re.sub(r"_+", "_", safe_id)
    # Remove leading and trailing underscores
    safe_id = safe_id.strip("_")

    filename = f"{class_name}-{safe_id or '_'}.{extension}"
    return filename


def get_current_async_loop():
    """
    get or automatically create an asnyc loop for the current thread.
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        # set_event_loop_policy() - why not?
        asyncio.set_event_loop(loop)
    return loop


class SerializableDataclass:
    """
    Presents uniform serialization for serializers using getstate and setstate and json
    serialization.
    """

    def json(self):
        return asdict(self)

    def __getstate__(self):
        return self.json()

    def __setstate__(self, values: typing.Dict):
        for key, value in values.items():
            setattr(self, key, value)


class Singleton(type):
    """enforces a Singleton"""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class MappableSingleton(Singleton):
    """Singleton with dict-like access to attributes"""

    def __setitem__(self, key, value) -> None:
        setattr(self, key, value)

    def __getitem__(self, key) -> typing.Any:
        return getattr(self, key)

    def __contains__(self, key) -> bool:
        return hasattr(self, key)


def get_input_model_from_signature(
    func: typing.Callable,
    remove_first_positional_arg: bool = False,
    ignore: typing.Sequence[str] | None = None,
    model_for_empty_annotations: bool = False,
) -> type[BaseModel] | None:
    """
    Create a pydantic model for a function's signature.

    Parameters
    ----------
    func: Callable
        The function for which to create the pydantic model.
    remove_first_positional_arg: bool, optional
        Remove the first argument from the model (this is appropriate for methods,
        as the first argument, self, is baked in when it's called, but is present
        in the signature).
    ignore: Sequence[str], optional
        Ignore arguments that have the specified name. This is useful for e.g.
        dependencies that are injected by LabThings.
    model_for_empty_annotations: bool, optional
        If True, create a model even if there are no annotations.

    Returns
    -------
    Type[BaseModel] or None
        A pydantic model class describing the input parameters, or None if there are no parameters.
    """
    parameters = OrderedDict(signature(func).parameters)  # type: OrderedDict[str, Parameter]
    if len(parameters) == 0:
        return None

    if all(p.annotation is Parameter.empty for p in parameters.values()) and not model_for_empty_annotations:
        return None

    if remove_first_positional_arg:
        name, parameter = next(iter((parameters.items())))  # get the first parameter
        if parameter.kind in (Parameter.KEYWORD_ONLY, Parameter.VAR_KEYWORD):
            raise ValueError("Can't remove first positional argument: there is none.")
        del parameters[name]

    # fields is a dictionary of tuples of (type, default) that defines the input model
    type_hints = typing.get_type_hints(func, include_extras=True)
    fields = {}  # type: typing.Dict[str, typing.Tuple[type, typing.Any]]
    for name, p in parameters.items():
        if ignore and name in ignore:
            continue
        if p.kind == Parameter.VAR_KEYWORD:
            p_type = typing.Annotated[
                typing.Dict[str, typing.Any] if p.annotation is Parameter.empty else type_hints[name],
                Parameter.VAR_KEYWORD,
            ]
            default = dict() if p.default is Parameter.empty else p.default
        elif p.kind == Parameter.VAR_POSITIONAL:
            p_type = typing.Annotated[
                typing.Tuple if p.annotation is Parameter.empty else type_hints[name],
                Parameter.VAR_POSITIONAL,
            ]
            default = tuple() if p.default is Parameter.empty else p.default
        else:
            # `type_hints` does more processing than p.annotation - but will
            # not have entries for missing annotations.
            p_type = typing.Any if p.annotation is Parameter.empty else type_hints[name]
            # pydantic uses `...` to represent missing defaults (i.e. required params)
            default = Field(...) if p.default is Parameter.empty else p.default
        fields[name] = (p_type, default)

    # If there are no fields, we don't want to return a model
    if len(fields) == 0:
        return None

    model = create_model(  # type: ignore[call-overload]
        f"{func.__name__}_input",
        **fields,
        __config__=ConfigDict(extra="forbid", strict=True, arbitrary_types_allowed=True),
    )
    return model


def get_return_type_from_signature(func: typing.Callable) -> RootModel | None:
    """Determine the return type of a function."""
    sig = inspect.signature(func)
    if sig.return_annotation == inspect.Signature.empty:
        return None  # type: ignore[return-value]
    else:
        # We use `get_type_hints` rather than just `sig.return_annotation`
        # because it resolves forward references, etc.
        type_hints = typing.get_type_hints(func, include_extras=True)
        from .core.property import wrap_plain_types_in_rootmodel

        if (
            "return" not in type_hints
            or type_hints["return"] is typing.Any
            or type_hints["return"] is None
            or type_hints["return"] is type(None)
        ):
            return None

        return wrap_plain_types_in_rootmodel(type_hints["return"])


def pydantic_validate_args_kwargs(
    model: typing.Type[BaseModel],
    args: typing.Tuple = tuple(),
    kwargs: typing.Dict = dict(),
) -> None:
    """
    Validate and separate *args and **kwargs according to the fields of the given pydantic model.

    Parameters
    ----------
    model: Type[BaseModel]
        The pydantic model class to validate against.
    *args: tuple
        Positional arguments to validate.
    **kwargs: dict
        Keyword arguments to validate.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the arguments do not match the model's fields.
    ValidationError
        If the arguments are invalid
    """

    field_names = list(model.model_fields.keys())
    data = {}

    # Assign positional arguments to the corresponding fields
    for i, arg in enumerate(args):
        if i >= len(field_names):
            raise ValueError(f"Too many positional arguments. Expected at most {len(field_names)}.")
        field_name = field_names[i]
        if Parameter.VAR_POSITIONAL in model.model_fields[field_name].metadata:
            if typing.get_origin(model.model_fields[field_name].annotation) is list:
                data[field_name] = list(args[i:])
            else:
                data[field_name] = args[i:]  # *args become end of positional arguments
            break
        elif field_name in data:
            raise ValueError(f"Multiple values for argument '{field_name}'.")
        data[field_name] = arg

    extra_kwargs = {}
    # Assign keyword arguments to the corresponding fields
    for key, value in kwargs.items():
        if key in data or key in extra_kwargs:  # Check for duplicate arguments
            raise ValueError(f"Multiple values for argument '{key}'.")
        if key in field_names:
            data[key] = value
        else:
            extra_kwargs[key] = value

    if extra_kwargs:
        for i in range(len(field_names)):
            if Parameter.VAR_KEYWORD in model.model_fields[field_names[i]].metadata:
                data[field_names[i]] = extra_kwargs
                break
            elif i == len(field_names) - 1:
                raise ValueError(f"Unexpected keyword arguments: {', '.join(extra_kwargs.keys())}")
    # Validate and create the model instance
    model.model_validate(data)


def json_schema_merge_args_to_kwargs(
    schema: dict, args: typing.Tuple = tuple(), kwargs: typing.Dict = dict()
) -> typing.Dict[str, typing.Any]:
    """
    Merge positional arguments into keyword arguments according to the schema.

    Parameters
    ----------
    schema: dict
        The JSON schema to validate against.
    args: tuple
        Positional arguments to merge.
    kwargs: dict
        Keyword arguments to merge.

    Returns
    -------
    dict
        The merged arguments as a dictionary, usually a JSON
    """
    if schema["type"] != "object":
        raise ValueError("Schema must be an object.")

    field_names = list(OrderedDict(schema["properties"]).keys())
    data = {}

    for i, arg in enumerate(args):
        if i >= len(field_names):
            raise ValueError(f"Too many positional arguments. Expected at most {len(field_names)}.")
        field_name = field_names[i]
        if field_name in data:
            raise ValueError(f"Multiple values for argument '{field_name}'.")
        data[field_name] = arg

    extra_kwargs = {}
    # Assign keyword arguments to the corresponding fields
    for key, value in kwargs.items():
        if key in data or key in extra_kwargs:  # Check for duplicate arguments
            raise ValueError(f"Multiple values for argument '{key}'.")
        if key in field_names:
            data[key] = value
        else:
            extra_kwargs[key] = value

    if extra_kwargs:
        data.update(extra_kwargs)
    return data


def get_all_sub_things_recusively(thing) -> typing.List:
    sub_things = [thing]
    for sub_thing in thing.sub_things.values():
        sub_things.extend(get_all_sub_things_recusively(sub_thing))
    return sub_things


def forkable(func):
    """
    Decorator to make a function forkable.
    This is useful for functions that need to be run in a separate thread.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        forked = kwargs.get("forked", False)  # Extract 'fork' argument, default to False
        if forked:
            thread = threading.Thread(target=func, args=args, kwargs=kwargs, daemon=True)
            thread.start()
            return thread
        else:
            return func(*args, **kwargs)

    return wrapper


def set_global_event_loop_policy():
    if sys.platform.lower().startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    from .config import global_config

    if global_config.USE_UVLOOP and sys.platform.lower() in [
        "linux",
        "darwin",
        "linux2",
    ]:
        import uvloop

        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


__all__ = [
    get_IP_from_interface.__name__,
    format_exception_as_json.__name__,
    pep8_to_dashed_name.__name__,
    get_default_logger.__name__,
    run_coro_sync.__name__,
    run_callable_somehow.__name__,
    complete_pending_tasks_in_current_loop.__name__,
    get_signature.__name__,
    isclassmethod.__name__,
    issubklass.__name__,
    get_current_async_loop.__name__,
    get_input_model_from_signature.__name__,
    pydantic_validate_args_kwargs.__name__,
    get_return_type_from_signature.__name__,
    getattr_without_descriptor_read.__name__,
    forkable.__name__,
    set_global_event_loop_policy.__name__,
]



from typing import Any
import logging
import uuid
import aiomqtt
import copy
import ssl


from ..utils import get_current_async_loop
from .utils import consume_broker_queue, consume_broker_pubsub_per_event
from ..config import global_config
from ..constants import Operations
from ..serializers import Serializers
from ..td.interaction_affordance import EventAffordance, PropertyAffordance
from ..param.parameters import Selector, String, Integer, ClassSelector, List


class MQTTPublisher:
    """
    MQTT Publisher for publishing messages to a specific topic.
    All events defined on the Thing will be published to MQTT topics with topic name "{thing id}/{event name}"

    For setting up an MQTT broker if one does not exist,
    see [infrastructure project](https://github.com/hololinked-dev/daq-system-infrastructure).
    """

    qos = Selector(objects=[0, 1, 2], default=0)
    """The MQTT QoS level to use for publishing messages"""

    hostname = String(default="localhost")
    """The MQTT broker hostname"""

    port = Integer(default=1883)
    """The MQTT broker port"""

    logger = ClassSelector(class_=logging.Logger, default=global_config.logger())  # type: logging.Logger
    """Logger instance"""

    things = List(default=None, allow_None=True, item_type=tuple)  # type: list[tuple[str, str, str]]
    """List of things to publish events from, each defined as a tuple of (server id, access point, thing id)"""

    ssl_context = ClassSelector(class_=ssl.SSLContext, allow_None=True, default=None)

    def __init__(self, hostname: str, port: int, username: str, password: str, qos: int = 1, **kwargs):
        """
        Parameters
        ----------
        hostname: str
            The MQTT broker hostname
        port: int
            The MQTT broker port
        username: str
            The MQTT broker username
        password: str
            The MQTT broker password
        qos: int
            The MQTT QoS level to use for publishing messages
        kwargs: dict
            Additional keyword arguments
        """
        self.hostname = hostname
        self.port = port
        self.qos = qos
        self.things = []
        self.username = username
        self.password = password
        self.logger = kwargs.get("logger", global_config.logger())
        self.ssl_context = kwargs.get("ssl_context", None)
        self._stop_publishing = False

    async def setup(self):
        """
        Sets up the MQTT client and starts publishing events from the configured things.
        All events are dispatched to their own async tasks. This method returns when setup is complete.
        """
        self._stop_publishing = False
        self.client = aiomqtt.Client(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
            tls_context=self.ssl_context,
        )
        try:
            await self.client.__aenter__()
        except aiomqtt.MqttReentrantError:
            pass
        for server_id, thing_id, access_point in self.things:
            thing, td = await consume_broker_queue(
                id=f"{self.hostname}:{self.port}|mqtt-publisher|{uuid.uuid4().hex[:8]}",
                server_id=server_id,
                thing_id=thing_id,
                access_point=access_point,
                logger=self.logger,
                context=global_config.zmq_context(),
            )
            eventloop = get_current_async_loop()
            for event_name in td.get("events", {}).keys():
                event_affordance = EventAffordance.from_TD(event_name, td)
                eventloop.create_task(self.publish(event_affordance))
            eventloop.create_task(self.publish_thing_description(td))

    async def publish(self, resource: EventAffordance):
        """
        Publishes an event to the MQTT broker.

        Parameters
        ----------
        resource: EventAffordance
            The event affordance for which the events are to be published
        """
        consumer = consume_broker_pubsub_per_event(resource)
        consumer.subscribe()
        topic = f"{resource.thing_id}/{resource.name}"
        while not self._stop_publishing:
            try:
                message = await consumer.receive()
                if message is None:
                    continue
                if message.body[1].value != b"":
                    payload = message.body[1]
                else:
                    payload = message.body[0]
                await self.client.publish(
                    topic=topic,
                    payload=payload.value,
                    qos=self.qos,
                    properties=dict(content_type=payload.content_type),
                )
            except Exception as ex:
                self.logger.error(f"Error publishing MQTT message: {ex}")

    def stop(self):
        """stop publishing, the client is not closed automatically"""
        self._stop_publishing = True

    def start(self):
        eventloop = get_current_async_loop()
        eventloop.create_task(self.setup())

    def add_thing(self, server_id: str, thing_id: str, access_point: str):
        """Adds a thing to the list of things to publish events from."""
        if access_point.upper() not in ["TCP", "IPC", "INPROC"]:
            raise ValueError("Access point must be 'TCP', 'IPC', or 'INPROC'")
        self.things.append((server_id, thing_id, access_point))

    async def publish_thing_description(self, ZMQ_TD: dict[str, Any]) -> dict[str, Any]:
        """Returns the Thing Description of the specified thing."""
        TD = copy.deepcopy(ZMQ_TD)
        # remove actions as they dont push events
        TD.pop("actions", None)
        # remove properties that are not observable
        for name in ZMQ_TD.get("properties", {}).keys():
            affordance = PropertyAffordance.from_TD(name, ZMQ_TD)
            if not affordance.observable:
                TD["properties"].pop(name)
                continue
            TD["properties"][name]["forms"] = []
            form = affordance.retrieve_form(Operations.observeproperty)
            form.href = f"mqtt{'s' if self.ssl_context else ''}://{self.hostname}:{self.port}"
            form.mqv_topic = f"{TD['id']}/{name}"
            TD["properties"][name]["forms"].append(form.json())
        # repurpose event
        for name in ZMQ_TD.get("events", {}).keys():
            affordance = EventAffordance.from_TD(name, ZMQ_TD)
            TD["events"][name]["forms"] = []
            form = affordance.retrieve_form(Operations.subscribeevent)
            form.href = f"mqtt{'s' if self.ssl_context else ''}://{self.hostname}:{self.port}"
            form.mqv_topic = f"{TD['id']}/{name}"
            TD["events"][name]["forms"].append(form.json())
        await self.client.publish(
            topic=f"{TD['id']}/thing-description",
            payload=Serializers.json.dumps(TD),
            qos=2,
            properties=dict(content_type="application/json"),
            retain=True,
        )



import inspect
import typing
from typing import ClassVar
from pydantic import BaseModel


class Schema(BaseModel):
    """
    Base dataclass for all WoT schema; Implements a custom asdict method which replaces dataclasses' asdict
    utility function
    """

    skip_keys: ClassVar = []  # override this to skip some dataclass attributes in the schema

    def model_dump(self, **kwargs) -> dict[str, typing.Any]:
        """Return the JSON representation of the schema"""
        # we need to override this to work with our JSON serializer
        kwargs["mode"] = "json"
        kwargs["by_alias"] = True
        kwargs["exclude_unset"] = True
        kwargs["exclude"] = [
            "instance",
            "skip_keys",
            "skip_properties",
            "skip_actions",
            "skip_events",
            "ignore_errors",
            "allow_loose_schema",
        ]
        return super().model_dump(**kwargs)

    def json(self) -> dict[str, typing.Any]:
        """same as model_dump"""
        return self.model_dump()

    @classmethod
    def format_doc(cls, doc: str):
        """strip tabs, newlines, whitespaces etc. to format the docstring nicely"""
        # doc_as_list = doc.split('\n')
        # final_doc = []
        # for index, line in enumerate(doc_as_list):
        #     line = line.lstrip('\n').rstrip('\n')
        #     line = line.lstrip('\t').rstrip('\t')
        #     line = line.lstrip('\n').rstrip('\n')
        #     line = line.lstrip().rstrip()
        #     if index > 0:
        #         line = ' ' + line # add space to left in case of new line
        #     final_doc.append(line)
        # final_doc = ''.join(final_doc)
        doc = inspect.cleandoc(doc)
        # Remove everything after "Parameters\n-----" if present (when using numpydoc)
        marker = "Parameters\n-----"
        idx = doc.find(marker)
        if idx != -1:
            doc = doc[:idx]
        doc = doc.replace("\n", " ")
        doc = doc.replace("\t", " ")
        doc = doc.lstrip().rstrip()
        return doc
